/**  
 Classe per a dibuixar l'escena 3D i els seus components.
 - Controla el flag que estableix si el punter es troba a la Zona 3D.
 - Gestiona el cas d'us on l'usuari pot moure l'escena  d'esquerra a dreta i de dalt a baix prement 
 el boto dret (a la zona 3D) i arrastrant el ratolí per tota la finestra.
 - Gestiona el cas d'us on l'usuari clica les etiquetes de l'esquerra de la gràfica.
 */
final class  Controlador3D {
  private ModelPrincipal mp = null;
  private Model3D model = null;
  private Vista3D vista = null;
  private BD1 BD = null;
  private PGraphics g = null;

  /**
   Mètode constructor, sol·licitat per Principal.
   */
  public Controlador3D(Principal principal, ModelPrincipal modelprincipal) { 
    //Objectes per a implementar l'arquitectura MODEL_VISTA_CONTROLADOR
    mp = modelprincipal;
    BD = new BD1();
    model = new Model3D();
    vista = new Vista3D(principal);
    // Objecte per a dibuixar.
    g = createGraphics(BD.ampleZona3D, BD.alsadaZona3D, P3D);
  } 

  /**
   Mètode per a dibuixar tots els components de l'escena 3D
   */
  public void pintaEscena3D() {
    g.beginDraw(); 
    g.background(255, 255, 255);
    //Posem la camara al seu lloc.
    controladorCamera.situaCamera(g);
    vista.pintaVista();
    g.endDraw(); 
    image(g, model.calculaOrigenX(), model.calculaOrigenY());
  }

  /**
   Mètodes per a establir el valor del flag que diu si el punter està en aquesta zona.
   */
  public boolean isZona3D() {
    return model.isZona3D;
  }
  public void setZona3D(boolean valor) {
    model.isZona3D= valor;
  }

  /**
   Mètode sol·licitat pel controladorPrincipal.
   CAS D'US: a la zona 3D, l'usuari mantè pulsat el botò dret i arrastra el punter.
   REQUISIT_1: activar el flag que permet modificar l'origen on dibuixem l'escena 3D.
   */
  public void  desplasaVista3D(boolean valor) {
    model.mouZona3D = valor;
  }

  /**
   Mètode sol·licitat pel controlador d'un label.
   CAS D'US: l'usuari ha clicat algun dels labels laterals de la gràfica.
   REQUISIT_1: mostrar o amagar: el pla SAGITAL sinó el FRONTAL sinó el VERTICAL.
   */
  public void pintaPla(String nom) {
    if (controladorGrafica.isSuperior(nom)) model.pintaPlaFRONTAL = !model.pintaPlaFRONTAL;
    else if (controladorGrafica.isCentral(nom)) model.pintaPlaSAGITAL = !model.pintaPlaSAGITAL;
    else if (controladorGrafica.isInferior(nom)) model.pintaPlaHORITZONTAL = !model.pintaPlaHORITZONTAL;
  }

  //BASE DADES//////////////////////////////////////////////////// ///////////////////////////
  /**
   Classe per a encapsular les variables de valor fix.
   */
  private final class BD1 {
    //Per a definir els limits de la ZONA 3D
    public  final int rigth =500;
    public  final int top =30;
    public  final int bottom =450;
    //DIMENSIONS buffer zona3D. Són més grans per tal de moure tota l'escena.
    public  final int ampleZona3D =  Principal.BDPrincipal.AMPLE/2+300;
    public  final int alsadaZona3D = Principal.BDPrincipal.ALSADA +325;
    //COORDENADES VISTA 3D
    public  final int x_Vista3D = -170;
    public  final int y_Vista3D = -200;
    //Variables relacionades amb l'IMU
    //MIDES IMU
    public  final int ample =  92;
    public  final int alsada = 6;
    public  final int fons =  66;
    //NOM IMATGES IMU
    public  final String img_top = "bno055.jpg";
    public  final String img_front ="pins.jpg";
    public  final int h_fletxa = 12;// hipotenusa del triangle que dibuixa la fletxa.
    public  final int c_fletxa = 4; //catet del triangle que dibuixa la fletxa.
    public  final int c_pla = 125; //longitud entre les arestes d'un pla.
    public  final int w_eix = 125;//longitud entre l'origen i el final d'un Eix.
    public  final float w_arc = 170f;//ample de l'arc.
    public  final float h_arc = 170f;//alçada de l'arc.
    ///////////////////////////////////////////////////////
    // llista Lletres i noms.
    public  final String  s_pitch =new String(Character.toChars(unhex("03C6")));
    public  final String  s_yaw =new String(Character.toChars(unhex("03C8")));
    public  final String s_roll = new String(Character.toChars(unhex("03A6")));
    public final int Id_N = 0; 
    public final int Id_S = 1;
    public final int Id_E = 2;
    public final int Id_O = 3;
    public final int Id_UP = 4;
    public final int Id_DONW = 5;
    public final int id_ESQUERRA = 6;
    public final int id_DRETA = 7;
    public final int id_DAVANT= 8;
    public final int id_DARRERE = 9;
    public final int id_SUPERIOR = 10;
    public final int id_INFERIOR= 11;
    public final PFont fontEscena= loadFont ("Dialog.plain-22.vlw");
    private final StringList noms =new StringList("N", "S", "E", "O", "UP", "DONW", 
      "Esquerra", "Dreta", "Davant", "Darrere", "Superior", "Inferior");
    public String getText(int id) {
      return noms.get(id);
    }
  }
  //MODEL//////////////////////////////////////////////////// ///////////////////////////
  /**
   Classe per a encapsular les variables que poden ser modificades.
   */
  private class Model3D {
    //Valors inicials.
    public boolean pintaPlaSAGITAL= false;
    public boolean pintaPlaHORITZONTAL= false;
    public boolean pintaPlaFRONTAL= false;
    public boolean mouZona3D= false;
    public int x_origen = BD.x_Vista3D;
    public int y_origen = BD.y_Vista3D;
    public boolean isZona3D = false;

    /**
     Mètodes per a calcular la posició de l'escena quan la movem.
     */
    public int calculaOrigenX() {
      if (mouZona3D)
        x_origen -= pmouseX - mouseX;
      return x_origen;
    }
    public int calculaOrigenY() {
      if (mouZona3D)
        y_origen -= pmouseY - mouseY;
      return y_origen;
    }
  }
  //VISTA//////////////////////////////////////////////////// ///////////////////////////
  private class Vista3D {
    private Box imu = null;

    /**
     Constructor, cridat pel Controlador.
     */
    public Vista3D(Principal principal) {
      //Instanciem un objecte BOX i, entre altres, li afegim les imatges.
      imu = new Box(BD.ample, BD.alsada, BD.fons);
      imu.rotateToX(PI/2);
      imu.drawMode(Box.SOLID, S3D.RIGHT | S3D.LEFT| S3D.BACK | S3D.BOTTOM)
        .fill(color(0, 0, 0), S3D.BOTTOM)
        .fill(color(100, 100, 100), S3D.BACK)
        .fill(color(150, 150, 150), S3D.RIGHT | S3D.LEFT);
      imu.use(S3D.TOP ).texture(principal, BD.img_top).drawMode(S3D.TEXTURE);
      imu.use(S3D.FRONT).texture(principal, BD.img_front).drawMode(S3D.TEXTURE);
    }

    /**
     Mètode per a dibuixar L'escena 3D.
     */
    public void pintaVista() {
      //Girem l'eix X per tal de situar el sistema de Processing igual al de l'IMU  
      g.rotateX(-PI/2);
      //Pintem el sistema fix.
      pintaSistemaFix();
      //Apliquem la matriu de rotació en funció de les dades rebudes.
      mp.setMatriuTransformacio(g);
      //Dibuixem tots els components de l'escena.
      pintaComponents();
      //Si cal dibuixar la Gravetat, deleguem la feina al controlador corresponent.
      controladorGravetat.pintaEscena(g);
    }

    private  void pintaSistemaFix() {
      g.textSize(14);
      g.strokeWeight(1);
      g.stroke(0, 0, 0); 
      g.fill(0, 0, 0); 
      //CAS D'US:  Amagar/mostrar els Eixos fixos.
      if (controladorPanel.isSeleccionat(controladorPanel.BD.EIXOS)) {
        //Pintem els eixos
        g.line(-BD.w_eix, 0, 0, BD.w_eix, 0, 0);//X
        g.line(0, -BD.w_eix, 0, 0, BD.w_eix, 0);//Y
        g.line(0, 0, -BD.w_eix, 0, 0, BD.w_eix);//Z
        pintaTextFix(135, 0, 0, BD.Id_O);//X
        pintaTextFix(-135, 0, 0, BD.Id_E);
        pintaTextFix(0, 135, 0, BD.Id_S);//Y
        pintaTextFix(0, -135, 0, BD.Id_N);
        pintaTextFix(0, 0, 140, BD.Id_DONW);//Z
        pintaTextFix(0, 0, -130, BD.Id_UP);
      }
      //CAS D'US:  Amagar/mostrar el perfil dels plans geogràfics.
      if (controladorPanel.isSeleccionat(controladorPanel.BD.B_PLANOLS)) {
        //CAS D'US:  pintar/no pintar els plans geogràfics.
        if (controladorPanel.isSeleccionat(controladorPanel.BD.PLANS_COLOR))
          //Pintem color negre amb transparencia del 40%
          g.fill(0, 0, 0, 40); 
        else g.noFill(); 
        g.stroke(0, 0, 0); 
        //Dibuixem els rectangles dels plans.
        g.rect(-BD.c_pla, -BD.c_pla, 2*BD.c_pla, 2*BD.c_pla );
        //Pla Vertical, cal fer una rotació.
        g.pushMatrix();
        g.rotateX(PI/2);
        g.rect(-BD.c_pla, -BD.c_pla, 2*BD.c_pla, 2*BD.c_pla );
        g.popMatrix();
      }
    }

    /**
     Mètode per a pintar el text del sistema fix i fer que miri sempre a l'observador.
     */
    private void pintaTextFix(float x, float y, float z, int idLletra) {
      g.pushMatrix();
      //Anem al lloc.
      g.translate(x, y, z);
      // Recuperem la posició original dels eixos.
      g.rotateX(PI/2);
      // Mirem de cara a l'observador.
      g.rotateY(radians(90-controladorCamera.getTheta()));
      g.rotateX(radians(- controladorCamera.getPhi()));
      //Pintem el text aproximadament centrada
      String s =  BD.getText(idLletra);
      g.text(s, -textWidth(s)/2, -2, 0);
      g.popMatrix();
    }

    /**
     Mètode per a pintar el text dels eixos de l'IMU (Anatòmics) i fer que miri sempre a l'observador.
     */
    private void pintaTextDinamic(float x, float y, float z, String s) {
      g.pushMatrix();
      g.translate(x, y, z);
      //Apliquem la matriu de rotació inversa.
      mp.setMatriuTransposada(g);
      // Recuperem la posició original dels eixos.
      g.rotateX(PI/2);
      // Mirem de cara a l'observador.
      g.rotateY(radians(90 - controladorCamera.getTheta()));
      g.rotateX(radians(- controladorCamera.getPhi())); 
      g.text(s, -textWidth(s)/2, -2, 0);
      g.popMatrix();
    }

    private void pintaComponents() {
      //CAS D'US:  pintar/no pintar el sistema d'eixos de l'IMU.
      if (controladorPanel.isSeleccionat(controladorPanel.BD.EIXOS_IMU)) {
        g.strokeWeight(3);
        g.fill(0, 0, 0);  
        g.textFont (BD.fontEscena, 10);
        pintaEixFrontal();
        pintaEixSagital();
        pintaEixVertical();
      }
      //CAS D'US:  pintar/no pintar l'IMU.
      if (controladorPanel.isSeleccionat(controladorPanel.BD.IMU)) 
        imu.draw(g);
      //CAS D'US:  pintar/no pintar els sectors dels angles de l'IMU.
      if (controladorPanel.isSeleccionat(controladorPanel.BD.ANGLES_IMU)) {
        g.strokeWeight(2);
        pintaAngleRoll(mp.roll);
        pintaAnglePitch(mp.pitch);
        pintaAngleYaw(mp.yaw);
      }
      pintaPla();
    }


    private void pintaEixFrontal() {
      g.stroke(255, 0, 0); 
      g.line(-110, 0, 0, 110, 0, 0);
      g.fill(0, 0, 0);
      pintaTextDinamic(110, 0, 0, BD.getText(BD.id_ESQUERRA));  // X
      pintaTextDinamic(-110, 0, 0, BD.getText(BD.id_DRETA));     // -X
    } 

    private  void pintaEixSagital() {
      g.stroke(0, 255, 100); 
      g.line(0, -110, 0, 0, 110, 0);
      g.fill(0, 0, 0);
      pintaTextDinamic(0, 110, 0, BD.getText(BD.id_DARRERE)); // Y
      pintaTextDinamic(0, -110, 0, BD.getText(BD.id_DAVANT));  // -Y
    }

    private  void pintaEixVertical() {
      g.stroke(0, 0, 255); 
      g.line(0, 0, 110, 0, 0, -110);
      g.fill(0, 0, 0);
      pintaTextDinamic(0, 0, -110, BD.getText(BD.id_SUPERIOR)); //-Z
      pintaTextDinamic(0, 0, 125, BD.getText(BD.id_INFERIOR));  // Z
    } 

    /**
     Mètode per a pintar el sector de l'angle ROLL
     */
    private void pintaAngleRoll(float angle) {
      g.pushMatrix();
      // Estem en el pla sagital, rotem Y per tal de canviar al pla frontal
      g.rotateX(PI/2);
      //Pintem de color negre
      g.stroke(0, 0, 0);
      g.fill(255, 0, 0);
      //Esquerra  Abducció – Adducció [0º, 90º)
      if (angle >= 0 && angle < 90) {
        //dibuixem  el sector de l'angle.
        g.arc(0.0, 0.0, BD.w_arc, -BD.h_arc, 0, radians(angle), PIE);
        g.rotateX(-PI/2);
        //ens traslladem al final de l'arc i pintem la punta de fletxa.
        g.translate(BD.w_arc/2, 0, 0);
        g.beginShape(TRIANGLE);
        g.vertex(0, 0, 0); 
        g.vertex(BD.c_fletxa, 0, -BD.h_fletxa);
        g.vertex(0, 0, -BD.h_fletxa);
        g.endShape();
        //Tornem enrrere, rotem i anem on volem pintar el text.
        g.translate(-BD.w_arc/2, 0, 0);
        g.rotateY(radians(angle/2));
        pintaTextDinamic(100, 0, 0, BD.s_roll);
      }
      //  Dreta  Abducció – Adducció [0º, -90º)
      else if (angle < 0 && angle > -90) {
        // //dibuixem  el sector de l'angle.
        g.arc(0.0, 0.0, -BD.w_arc, BD.h_arc, radians(angle), 0, PIE);
        g.rotateX(-PI/2);
        //ens traslladem al final de l'arc i pintem la punta de fletxa.
        g.translate(-BD.w_arc/2, 0, 0);
        //Pintem la mitat d'una fletxa
        g.beginShape(TRIANGLE);
        g.vertex(0, 0, 0); 
        g.vertex(-BD.c_fletxa, 0, -BD.h_fletxa);
        g.vertex(0, 0, -BD.h_fletxa);
        g.endShape();
        //Tornem enrrere, rotem i anem on volem pintar el text.
        g.translate(BD.w_arc/2, 0, 0);
        g.rotateY(radians(angle/2));
        pintaTextDinamic(-100, 0, 0, BD.s_roll);
      }
      g.popMatrix();
    }
    /**
     Mètode per a pintar el sector de l'angle PITCH
     */
    private void pintaAnglePitch(float angle) {
      g.pushMatrix();
      g.stroke(0, 0, 0);
      g.fill(0, 255, 0);
      g.rotateY(-PI/2);
      //dibuixem  el sector
      //Ventral  Flexió – Extensió (0º, 180º)
      if (angle > 0 && angle < 180)
        g.arc(0.0, 0.0, -BD.w_arc, BD.h_arc, 0, radians(angle), PIE);
      //Dorsal  Flexió – Extensió (0º, -180º)
      else if (angle < 0 && angle > -180)
        g.arc(0.0, 0.0, -BD.w_arc, BD.h_arc, radians(angle), 0, PIE);
      g.rotateX(-PI/2);
      //ens traslladem al final de l'arc i pintem la punta de fletxa.
      g.translate(-BD.w_arc/2, 0, 0);
      g.beginShape(TRIANGLE);
      g.vertex(0, 0, 0); 
      if (angle > 0 && angle < 180) {
        g.vertex(-BD.c_fletxa, 0, BD.h_fletxa);
        g.vertex(0, 0, BD.h_fletxa);
      } else if (angle < 0 && angle > -180) {
        g.vertex(-BD.c_fletxa, 0, -BD.h_fletxa);
        g.vertex(0, 0, -BD.h_fletxa);
      }
      g.vertex(0, 0, -BD.h_fletxa);
      g.endShape();
      //Tornem enrrere, rotem i anem on volem pintar el text.
      g.translate(BD.w_arc/2, 0, 0);
      g.rotateX(-PI/2);
      g.rotateY(-PI/2);
      g.rotateZ(-PI/2);
      g.rotateY(radians(angle/2));
      g.fill(200, 255, 200);
      pintaTextDinamic(0, 0, 100, BD.s_pitch);
      g.popMatrix();
    }

    /**
     Mètode per a pintar el sector de l'angle YAW
     */
    private void pintaAngleYaw(float angle) {
      g.pushMatrix();
      g.stroke(0, 0, 0);
      g.fill(0, 0, 255);
      g.pushMatrix();
      g.rotateZ(-PI/2);
      g.arc(0.0, 0.0, BD.w_arc, BD.h_arc, 0, radians(angle), PIE);
      g.translate(85.0, 0, 0);
      g.rotateX(-PI/2);
      //Pintem la mitat d'una fletxa
      g.beginShape(TRIANGLE);
      g.vertex(0, 0, 0); 
      g.vertex(BD.c_fletxa, 0, BD.h_fletxa);
      g.vertex(0, 0, BD.h_fletxa);
      g.endShape();
      g.popMatrix();
      //Ens desplacem al llarg del perimetre per tal de dibuixar la lletra.
      g.rotateZ(radians(20));
      pintaTextDinamic(0, -100, 0, BD.s_yaw);
      g.popMatrix();
    }
    private void pintaPla() {
      //Els moviments en el pla Frontal són: 
      //  Abducció → separació de l'eix vertical.
      //  Adducció → aproximació a l’eix vertical..
      //CAS D'US:  pintar/no pintar el pla FRONTAL.
      g.strokeWeight(1);
      if (model.pintaPlaFRONTAL) {
        g.pushMatrix();
        g.rotateX(-PI/2);
        g.fill(255, 0, 0, 75);
        g.rect(-BD.c_pla, -BD.c_pla, 2*BD.c_pla, 2*BD.c_pla);
        g.popMatrix();
      }
      //Els moviments en el pla Sagital són: 
      //   Flexió → disminució de l'angle pitch entre dos ossos o zones 
      //   Extensió → augment de l'angle pitch entre dos ossos o zones. 
      //CAS D'US:  pintar/no pintar el pla SAGITAL.
      if (model.pintaPlaSAGITAL) {
        g.pushMatrix();
        g.rotateY(PI/2);
        g.fill(0, 255, 0, 75);
        g.rect(-BD.c_pla, -BD.c_pla, 2*BD.c_pla, 2*BD.c_pla);
        g.popMatrix();
      }
      //CAS D'US:  pintar/no pintar el pla HORITZONTAL.
      if (model.pintaPlaHORITZONTAL) {
        g.pushMatrix();
        g.fill(0, 0, 255, 75);
        g.rect(-BD.c_pla, -BD.c_pla, 2*BD.c_pla, 2*BD.c_pla);
        g.popMatrix();
      }
    }
  }
}
