/**  //<>//
 Classe per a  implementar els moviments de l'observador. 
 - Gestiona el cas d'us on l'usuari pot apropar o allunyar la posició  de  l'observador,
 prement el boto central i arrossegant el ratolí de costat a costat .
 - Gestiona el cas d'us on l'usuari pot desplaçar la posició  de  l'observador,
 per l'escena desplaçant-se per la superfície d'una esfera que  la conté. Per a fer-ho, 
 ha de situar el punter en la zona 3D, prémer el botó esquerra i arrossegar el punter. 
 - Gestiona el cas d'us on l'usuari pot moure la posició  de  l'observador,
 amb les rodes que regulen els angles phi (angle vertical) i theta (angle horitzontal).
 */
public final class ControladorCamera {
  private ModelCamera model = null;
  private VistaCamera vista = null;
  private BD1 BD = null;

  /**
   Mètode constructor, sol·licitat per Principal.
   */
  public ControladorCamera() { 
    //Objectes per a implementar l'arquitectura MODEL_VISTA_CONTROLADOR
    BD = new BD1();
    model = new ModelCamera();
    vista = new VistaCamera();
  } 

  /**
   Mètode sol·licitat pel controladorPrincipal.
   CAS D'US: L'usuari prem el boto esquerre del ratolì  i arrastra el punter per la zona3D
   REQUISIT_1: Calcular la nova posició de la camara.
   REQUISIT_2: Actualitzar el valor que mostren les vistes de les rodes.
   */
  public void rodarCamara(int xmouse, int ymouse) {
    model.calculaPosicio(xmouse, ymouse);
    vista.r_phi.setValor((int)getPhi(), BD.s_grau);
    vista.r_theta.setValor((int)getTheta(), BD.s_grau);
  }

  /**
   Mètode sol·licitat pel controladorPrincipal.
   CAS D'US: L'usuari deixà de pulsar el ratolì, després d'arrosegar-lo.
   REQUISIT_1: Desar els valors dels angles.
   */
  public void desarAnglesCamera() {
    model.capturaAngles();
  }
  /**
   Mètode sol·licitat pel controladorRoda.
   CAS D'US:  L'usuari se situa damunt la roda que regula l'angle Theta de la càmera i fa rodar el boto central del ratolì.
   REQUISIT_1: Calcular el nou valor de l'angle.
   REQUISIT_2: Actualitzar el valor de la vista de la roda.
   */
  public void rodaAngleTheta(int direccio) {
    model.calculaTheta(direccio);
    vista.r_theta.setValor((int)getTheta(), BD.s_grau);
  }
  /**
   Mètode sol·licitat pel controladorRoda.
   CAS D'US: L'usuari se situa damunt la roda que regula l'angle Phi de la càmera i fa rodar el boto central del ratolì.
   REQUISIT_1: Calcular el nou valor de l'angle.
   REQUISIT_2: Actualitzar el valor de la vista de la roda.
   */
  public void rodaAnglePhi(int direccio) {
    model.calculaPhi(direccio);
    vista.r_phi.setValor((int)getPhi(), BD.s_grau);
  }
  /**
   Mètode sol·licitat pel controlador3D.
   CAS D'US: pintar l'escena 3D.
   REQUISIT_1: Establir la posició de la càmera configurant la posició dels ulls, 
   el centre de l'escena i quin eix està orientat cap amunt.
   */
  public void  situaCamera( PGraphics g) {
    model.calculaXYZ();
    g.camera(model.ull_X, model.ull_Y, model.ull_Z, 
      0, 0, 0, 
      -sin(model.v)* cos(model.h), cos(model.v), -sin(model.v)* sin(model.h));
  } 

  public float getTheta() {
    return  model.theta_c;
  }
  public float getPhi() {
    return  model.phi_c;
  }
  /**
   Mètode sol·licitat pel controladorPrincipal.
   CAS D'US:  En la zona 3D, l'usuari mantè pulsat el botò.
   REQUISIT_1: Capturar la posició x, y del punter.
   */
  public void capturaMouse(int x, int y) {
    model.xPresionat = x;
    model.yPresionat = y;
  }
  /**
   Mètode sol·licitat pel controladorPrincipal.
   CAS D'US: Amb el punter situat a la zona3D, premer el botò central i arrosegar el punter de costat a costat.
   REQUISIT_1: Incrementar o decrementar la posició de l'observador.
   */
  public void mouObservador(int desplasament) {
    if (desplasament > 0 ) model.modulObservador -=1;
    if (desplasament < 0) model.modulObservador+=1;
  }
  //BASE DADES//////////////////////////////////////////////////// ///////////////////////////
  /**
   Classe per a encapsular les variables de valor fix.
   */
  private final class BD1 {
    public final String  s_theta = new String(Character.toChars(unhex("03B8"))); 
    public  final String  s_phi = new String(Character.toChars(unhex("03C6")));
    //Constants per a indexar les rodes .
    public  final int  id_r_theta = 55;
    public  final int  id_r_phi = 66;
    public  final String  valor_r_phi = " "+s_phi +" "+"PHI ";
    public  final String  valor_r_theta =" "+s_theta +" "+"THETA ";
    public String nom_r_theta = "camera_theta";
    public String nom_r_phi = "camera_phi" ;
    public  final String  s_grau = "º";
    //Dimensions Rodes
    public  final int  ample = 7;
    public  final int  alsada = 10;
  }

  //MODEL//////////////////////////////////////////////////// ///////////////////////////
  private class ModelCamera {
    private float ull_X, ull_Y, ull_Z;
    private float phi_c, theta_c, v, h ;
    private float xPresionat, yPresionat, xUltim, yUltim; 
    private  int modulObservador;

    public ModelCamera() {
      //posició inicial
      phi_c= -45;
      theta_c =-120;
      modulObservador =550;
      v =radians(phi_c);
      h = radians(theta_c);
      xUltim = theta_c; 
      yUltim = phi_c;
      ull_X = ull_Y = ull_Z = 0.0;
    }

    /**
     Mètode per a calcular les coordenades de l'observador
     en funció dels angles phi i theta.
     */
    public void calculaXYZ() {
      ull_X= modulObservador * cos(v)* cos(h);
      ull_Y = modulObservador * sin(v);
      ull_Z = modulObservador * cos(v)  *sin(h);
    }
    public void calculaPosicio(int xmouse, int ymouse) {
      theta_c = -xmouse + xPresionat + xUltim;
      phi_c = ymouse - yPresionat + yUltim;
      if (theta_c < -360)  theta_c = theta_c +360;
      else if (theta_c >360) theta_c = theta_c -360;
      if (phi_c < -360)  phi_c = phi_c +360;
      else if (phi_c >360) phi_c = phi_c -360;
      v = radians(phi_c);
      h = radians(theta_c);
    }

    public void calculaTheta(float increment) {
      theta_c = (theta_c +increment) % 360;
      h = radians(theta_c);
      //desem l'ultima posició.
      xUltim = theta_c ;
    }

    public void calculaPhi(float increment) {
      phi_c = (phi_c+increment) % 360;
      v =radians(phi_c);
      //desem l'ultima posició.
      yUltim = phi_c ;
    }
    public void capturaAngles() {
      xUltim = theta_c ;
      yUltim = phi_c ;
    }
  }
  //VISTA///////////////////////////////////////////////////////// ///////////////////////////
  private class VistaCamera {
    public ControladorRoda r_phi, r_theta ;

    public VistaCamera() {
      creaRodaPhi();
      creaRodaTheta();
    }
    public void creaRodaPhi() {
      int x = Principal.BDPrincipal.AMPLE/4-60;
      int y = Principal.BDPrincipal.ALSADA-60;
      r_phi= new ControladorRoda(x, y, BD.ample, BD.alsada, BD.id_r_phi, BD.nom_r_phi);
      r_phi.setRange(-360, 360);
      r_phi.setValor((int)model.phi_c, BD.s_grau);
      setVista(r_phi, BD.valor_r_phi);
    }
    public void creaRodaTheta() {
      int x = Principal.BDPrincipal.AMPLE/4;
      int y = Principal.BDPrincipal.ALSADA - 60;
      r_theta= new ControladorRoda(x, y, BD.ample, BD.alsada, BD.id_r_theta, BD.nom_r_theta);
      r_theta.setRange(-360, 360);
      r_theta.setValor((int)model.theta_c, BD.s_grau);
      setVista(r_theta, BD.valor_r_theta);
    }
    public void setVista( Numberbox roda, String nom) {
      roda. getCaptionLabel().set(nom);
      roda.getCaptionLabel().setVisible(true);
      roda.getCaptionLabel().setColor(ControlP5.BLACK);
      roda.getCaptionLabel().setColorBackground(0xFFFFFF00);
    }
  }
}
