/** 
 Classe per a representar les dades.
 - Gestiona el cas d'us on l'usuari clica  un item de la llista desplegable.
 - Gestiona el cas d'us on l'usuari interacciona amb les rodes de les escales del panell esquerra de la gràfica.
 - Gestiona el cas d'us on l'usuari modifica l'`ltim registre representat a pantalla.
 - Gestiona el cas d'us on l'usuari interacciona amb la roda que regula la velocitat de dibuix.
 */
public final class ControladorGrafica {
  private ModelPrincipal mp = null;
  private ModelGrafica model = null;
  private VistaGraficaLateral vista = null;
  private VistaGraficaSD vista_grafica=null;
  public BD1 BD = null;

  /**
   Mètode constructor, sol·licitat per Principal.
   */
  public ControladorGrafica(ModelPrincipal modelPrincipal) { 
    //Objectes per a implementar l'arquitectura MODEL_VISTA_CONTROLADOR
    BD = new BD1();
    mp = modelPrincipal;
    model = new ModelGrafica();
    vista = new VistaGraficaLateral();
  }

  /**
   Mètode cridat per Principal.
   Si tenim dades pel port, instancia la vista de la gràfica i inicia la taula de treball.
   Si no tenim dades al port, estableix els fotogrmes per segon i instancia la vista de la gràfica per 
   a treballar amb les dades d'un arxiu.
   */
  public void init(boolean portOk) {
    //Iniciem la gáfica.
    if (portOk) {
      surface.setTitle("DISPOSIITU EN TEMPS REAL");
      model.tenimPort = portOk;
      vista_grafica = new VistaGrafica(this);
      mp.iniciaTaula();
    } else {
      //Establim la quantitat de frems que pintem en un segon.
      frameRate(BD.fremsps);
      vista_grafica = new VistaGraficaSD(this);
      vista_grafica.init();
    }
  }

  /**
   Mètode sol·licitat pel controladorPrincipal.
   CAS D'US: L'usuari ha clicat un item de la llista desplegable.
   REQUISIT_1: Carregar l'arxiu referenciat per l'id del ítem a la taula de treball.
   */
  public void obrirArxiu(int idItem) {
    String nomArxiu = (String)vista_grafica.ls.getItem(idItem).get("name");
    surface.setTitle("DISPOSIITU >>>>>>>>>>>>>>>>>> "+  split(nomArxiu, '.')[0]);
    mp.iniciaTaula(nomArxiu);
    model.tenimArxiu = true;
    //Si estem en pause, pintem la nova situació.
    if (isPause()) loop();
  }

  /**
   Mètode sol·licitat pel controladorPrincipal.
   Gestiona el dibuixat de l'escena.
   */
  public void pintaEscena2D() {
    //Mirem si estem treballant amb les dades d'un arxiu.
    if (model.tenimArxiu) {
      //Quan fem pause, pintem fins aqui i parem.
      if (isPause()) noLoop();
      //Per defecte, cada 8 frame pintem un registre nou.
      if (frameCount % model.v_fotograma ==0 )
        mp.avansaIDRegistre();
      vista_grafica.pintaEscena();
    } else if (model.tenimPort)  ((VistaGrafica)vista_grafica).pintaEscena();
    //Demanem al controlador que pinti
    controladorPanel.pintaCaixes();
  }

  /**
   Mètode sol·licita pel controladorRoda.
   CAS D'US: L'usuari situa el punter a sobre de la representació gràfica d'una de les rodes de la zona
   esquerre de la gràfica i fa rodar el botó central del ratolí.
   REQUISIT_1: modificar el valor de la roda i representar la nova situació.
   */
  public void mouLimits(int increment, int idRoda) {
    if (idRoda ==BD.idr_roll_d) { //Només movem la roda si estem per sota del valor máxim permés.
      if ( mp.roll <= model.roll_d_escala) {
        model.roll_d_escala += increment;
        if (model.roll_d_escala > vista.rd_roll.getMax()) model.roll_d_escala = (int) vista.rd_roll.getMax() ;
        else if (model.roll_d_escala < 0 ) model.roll_d_escala =0;
        setValorVisual(vista.rd_roll, model.roll_d_escala);
      }
    }
    if (idRoda ==BD.idr_roll_e) { //Només movem la roda si estem per sota del valor máxim permés.
      if ( mp.roll> model.roll_e_escala) {
        model.roll_e_escala += increment;
        if (model.roll_e_escala < vista.re_roll.getMin()) model.roll_e_escala = (int) vista.re_roll.getMin() ;
        else if (model.roll_e_escala > 0 ) model.roll_e_escala =0;
        setValorVisual(vista.re_roll, model.roll_e_escala);
      }
    }
    if (idRoda ==BD.idr_pitch_r) { ////Només movem la roda si estem per sota del valor máxim permés.
      if ( mp.pitch <= model.pitch_a_escala) {
        model.pitch_a_escala += increment;
        if (model.pitch_a_escala > vista.ra_pitch.getMax()) model.pitch_a_escala = (int) vista.ra_pitch.getMax() ;
        else if (model.pitch_a_escala < 0 ) model.pitch_a_escala =0;
        setValorVisual(vista.ra_pitch, model.pitch_a_escala);
      }
    }
    if (idRoda ==BD.idr_pitch_p) {// //Només movem la roda si estem per sota del valor máxim permés.
      if ( mp.pitch> model.pitch_p_escala) {
        model.pitch_p_escala += increment;
        if (model.pitch_p_escala < vista.rp_pitch.getMin()) model.pitch_p_escala = (int) vista.rp_pitch.getMin() ;
        else if (model.pitch_p_escala > 0 ) model.pitch_p_escala =0;
        setValorVisual(vista.rp_pitch, model.pitch_p_escala);
      }
    }
  }
  public  void setValorVisual(Numberbox roda, int valor) {
    roda.getValueLabel().setText(valor+BD.s_grau);
  }
  public float getYMitjaZonaSuperior() {
    return mp.getYMitja(2, model.idRowIniciMitja);
  }
  public float getYMitjaZonaCentre() {
    return mp.getYMitja(3, model.idRowIniciMitja);
  }
  public float getYMitjaZonaInferior() {
    return mp.getYMitja(1, model.idRowIniciMitja);
  }

  /**
   Mètode per a calcular la coordenada y del punt del valor corresponent a la zona superior
   del gràfic.
   */
  public float getYPuntZonaSuperior() {
    float y=0.0, aux, v;  
    if (controladorPanel.isSeleccionat(controladorPanel.BD.GRAFICA_EULER)) {
      if (mp.roll<0) {// Zona negativa.
        //Primer, mirem si cal ajustar l'escala del gràfic.
        if (mp.roll < model.roll_e_escala && mp.roll > vista.re_roll.getMin() ) {
          model.roll_e_escala = (int)mp.roll;
          setValorVisual(vista.re_roll, model.roll_e_escala);
        }
        //Ara fem la funció map amb els valors que ens interesen per tal d'obtenir la coordenada Y.
        aux = map(mp.roll, 0, model.roll_e_escala, BD.y_ZSuperiorZero, BD.y_ZSuperiorBottom);
        y = aux >= BD.y_ZSuperiorBottom ? BD.y_ZSuperiorBottom  : aux ;
      }      //hem modificat l'escala amb la roda, pero hi ha valors més grans, per tatn, canviem el valor del máxim.
      else {//zona positiva
        if (mp.roll > model.roll_d_escala && mp.roll < vista.rd_roll.getMax() ) {
          model.roll_d_escala = (int)mp.roll;
          setValorVisual(vista.rd_roll, model.roll_d_escala);
        }
        aux =map(mp.roll, model.roll_d_escala, 0, BD.y_ZSuperiorTop, BD.y_ZSuperiorZero);
        y = aux < BD.y_ZSuperiorTop ? BD.y_ZSuperiorTop : aux;
      }
    } else if (controladorPanel.isSeleccionat(controladorPanel.BD.GRAFICA_GRAVETAT)) {
      v = mp.ox/10.0;
      if (v<0) {// Zona negativa.
        //Primer, mirem si cal ajustar l'escala del gràfic.
        if (v < model.ox_n_escala && v> vista.rn_ox.getMin() ) {
          model.ox_n_escala = (int)v;
          setValorVisual(vista.rn_ox, model.ox_n_escala);
        }
        //Ara fem la funció map amb els valors que ens interesen per tal d'obtenir la coordenada Y.
        aux = map(v, 0, model.ox_n_escala, BD.y_ZSuperiorZero, BD.y_ZSuperiorBottom);
        y = aux >= BD.y_ZSuperiorBottom ? BD.y_ZSuperiorBottom  : aux ;
      }//hem modificat l'escala amb la roda, pero hi ha valors més grans, per tant, canviem el valor del máxim.
      else {//zona positiva
        if (v > model.ox_p_escala && v < vista.rp_ox.getMax() ) {
          model.ox_p_escala = (int)v;
          setValorVisual(vista.rp_ox, model.ox_p_escala);
        }
        aux =map(v, model.ox_p_escala, 0, BD.y_ZSuperiorTop, BD.y_ZSuperiorZero);
        y = aux < BD.y_ZSuperiorTop ? BD.y_ZSuperiorTop : aux;
      }
    }
    return y;
  }

  /**
   Mètode per a calcular la coordenada y del punt del valor corresponent a la zona Central
   del gràfic.
   */
  public float getYPuntZonaCentral() {
    float y=0.0, aux, v;   
    if (controladorPanel.isSeleccionat(controladorPanel.BD.GRAFICA_EULER)) {
      if (mp.pitch<0) {// Zona negativa.
        if (mp.pitch < model.pitch_p_escala && mp.pitch > vista.rp_pitch.getMin() ) {
          model.pitch_p_escala = (int)mp.pitch;
          setValorVisual(vista.rp_pitch, model.pitch_p_escala);
        }
        aux = map(mp.pitch, 0, model.pitch_p_escala, BD.y_ZCentreZero, BD.y_ZCentreBottom);
        y = aux >= BD.y_ZCentreBottom ? BD.y_ZCentreBottom : aux;
      } //hem modificat l'escala amb la roda, pero hi ha valors més grans, per tant, canviem el valor del máxim.
      else {//zona positiva
        if (mp.pitch > model.pitch_a_escala && mp.pitch < vista.ra_pitch.getMax() ) {
          model.pitch_a_escala = (int)mp.pitch;
          setValorVisual(vista.ra_pitch, model.pitch_a_escala);
        }
        aux =map(mp.pitch, model.pitch_a_escala, 0, BD.y_ZCentreTop, BD.y_ZCentreZero);
        y = aux < BD.y_ZCentreTop? BD.y_ZCentreTop : aux;
      }
    } else if (controladorPanel.isSeleccionat(controladorPanel.BD.GRAFICA_GRAVETAT)) {
      v = mp.oy/10.0;
      if (v<0) {// Zona negativa.
        if (v< model.oy_n_escala && v > vista.rn_oy.getMin() ) {
          model.oy_n_escala = (int)v;
          setValorVisual(vista.rn_oy, model.oy_n_escala);
        }
        aux = map(v, 0, model.oy_n_escala, BD.y_ZCentreZero, BD.y_ZCentreBottom);
        y = aux >= BD.y_ZCentreBottom ? BD.y_ZCentreBottom : aux;
      }//hem modificat l'escala amb la roda, pero hi ha valors més grans, per tant, canviem el valor del máxim.
      else {//zona positiva
        if (v > model.oy_p_escala && v < vista.rp_oy.getMax() ) {
          model.oy_p_escala = (int)v;
          setValorVisual(vista.rp_oy, model.oy_p_escala);
        }
        aux =map(v, model.oy_p_escala, 0, BD.y_ZCentreTop, BD.y_ZCentreZero);
        y = aux < BD.y_ZCentreTop? BD.y_ZCentreTop : aux;
      }
    }
    return y;
  }

  /**
   Mètode per a calcular la coordenada y del punt del valor corresponent a la zona Inferior
   del gràfic.
   */
  public float getYPuntZonaInferior() {
    float y=0.0, aux, v;  
    if (controladorPanel.isSeleccionat(controladorPanel.BD.GRAFICA_EULER)) {
      //zona positiva
      if (mp.yaw > model.yaw_escala && mp.yaw < vista.r_yaw.getMax() ) {
        model.yaw_escala = (int)mp.yaw;
        setValorVisual(vista.r_yaw, model.yaw_escala);
      }
      y = map(mp.yaw, model.yaw_escala, 0, BD.y_ZInferiorTop, BD.y_ZInferiorBottom);
    } else if (controladorPanel.isSeleccionat(controladorPanel.BD.GRAFICA_GRAVETAT)) {
      v = mp.oz/10.0;
      if (v<0) {// Zona negativa.
        if (v < model.oz_n_escala && v > vista.rn_oz.getMin() ) {
          model.oz_n_escala = (int)v;
          setValorVisual(vista.rn_oz, model.oz_n_escala);
        }
        aux = map(v, 0, model.oz_n_escala, BD.y_ZInferiorZero, BD.y_ZInferiorBottom);
        y = aux >= BD.y_ZInferiorBottom ? BD.y_ZInferiorBottom : aux;
      }      //hem modificat l'escala amb la roda, pero hi ha valors més grans, per tatn, canviem el valor del máxim.
      else {//zona positiva
        if (v > model.oz_p_escala && v < vista.rp_oz.getMax() ) {
          model.oz_p_escala = (int)v;
          setValorVisual(vista.rp_oz, model.oz_p_escala);
        }
        aux =map(v, model.oz_p_escala, 0, BD.y_ZInferiorTop, BD.y_ZInferiorZero);
        y = aux < BD.y_ZInferiorTop? BD.y_ZInferiorTop : aux;
      }
    }
    return y;
  }

  ////////////////////// Mètodes auxiliars.
  public void pintaValorLabelTemps() {
    vista.l_temps.l.setStringValue(mp.hora);
  }
  public void resetValorsLabelsLateral() {
    vista.setValorsLabels();
  }
  public void avansa() {
    model.xDial += BD.w_dial;
    model.idRegistre_pintar++;
  }
  public void getRegistre() {
    //Demanem a Model que carregui les dades del registre que volem pintar.
    mp.carregaRegistre(model.idRegistre_pintar);
  }
  public void getUltimRegistre() {
    //Demanem a Model que carregui les dades del registre que volem pintar.
    mp.carregaRegistre(model.idRowUltim);
  }
  public boolean isUltimRegistre() {
    boolean b = model.idRegistre_pintar == mp.idRegistre;
    // .. capturem l'id del registre i la posició del dial, 
    if (b) model.capturaDial();
    return b;
  }
  public boolean isRegistrePause() {
    if (model.isZonaGrafica ) {
      if (model.isPosicioDial()) {
        //... sempre capturem el registre que marca el punter.
        model.xMouse = model.xDial;
        model.idRowClicat =model.idRegistre_pintar;
        // i pintem el registre com si fos l'actual.
        controlador3D.pintaEscena3D();
        if ((model.idRegistre_pintar+1) % 4 ==0) 
          pintaValorLabelTemps();
        return true;
      }
    }
    return false;
  }
  ///////////////////////////////////////////////////////
  /**
   Mètode sol·licita pel controladorPrincipal.
   CAS D'US: l'animació de la gràfica està en pausa, el punter està a la zona3D  i l'usuari fa rodar el botó central.
   REQUISIT_1: modificar l'identificador de l’últim registre pintat i dibuixar la nova situació.
   */
  public void mouPunt(int direccio) {
    //Acció principal: Calculem el desplaçament.
    int id= direccio + model.idRowUltim;
    // Sub cas_1: Anem enrrere i el desplaçament és negatiu.
    //Acció: mostrar registres des de l'ultim desplaçament.
    if (id < 0 ) id=mp.table.lastRowIndex()-1;
    // Sub cas_2: Anem endavant i superem el nombre de registres.
    //Acció: posar a zero el desplaçament .
    else if (id > mp.table.lastRowIndex()) id=0;
    // Sub cas_3: Anem endavant i superem l'últim registre pintat,.
    //Acció: llegir un nou registre,
    else if (id > mp.idRegistre) 
      mp.carregaRegistre(id);
    // Sub cas_4: El registre correspon a un interval de temps equivalent al segon.
    //Acció: mostrar l'hora en la que s'ha fet el registre.
    if ((id+1) % 4 ==0) pintaValorLabelTemps();
    //Resta d'Accions: actualitzar variables involucrades.
    mp.idRegistre = id;
    model.idRowUltim = id;
    model.xUltim = BD.x_inici_dials + BD.w_dial*(id % BD.maxDials);
  }

  /**
   Mètode sol·licita pel controladorRoda.
   CAS D'US: el punter està a sobre de la roda que regula la velocitat de dibuix de la gràfica
   i l'usuari fa rodar el botó central del ratolí.
   REQUISIT_1: modificar el valor de la roda.
   */
  public void modificaFps(int direccio) {
    model.v_fotograma +=direccio;
    if (model.v_fotograma < 1 )model.v_fotograma = 1;
    else if (model.v_fotograma > BD.fremsps)model.v_fotograma = BD.fremsps;
    vista_grafica.getRodaVelocitat().setValor(model.v_fotograma, BD.s_fps);
  }

  ////////////////////// Mètodes auxiliars.
  public ModelGrafica getModel() {
    return model;
  }
  public BD1 getBD() {
    return new BD1();
  }
  public Boolean isSuperior(String nom) {
    return nom.equals(BD.getValorVisual(BD.idl_superior));
  }
  public Boolean isCentral(String nom) {
    return nom.equals(BD.getValorVisual(BD.idl_centre));
  }
  public Boolean isInferior(String nom) {
    return nom.equals(BD.getValorVisual(BD.idl_inferior));
  }

  public boolean isZonaGrafica() {
    return model.isZonaGrafica;
  }
  public void setZonaGrafica(boolean valor) {
    model.isZonaGrafica= valor;
  }
  public boolean isPause() {
    return model.pause;
  }
  public void setPause(boolean valor) {
    model.pause = valor;
  }
  public boolean isSeleccionatSuperior() {
    return(model.taula_yselec_superior.hasValue(model.idRegistre_pintar));
  }
  public boolean isSeleccionatCentre() {
    return(model.taula_yselec_centre.hasValue(model.idRegistre_pintar));
  }
  public boolean isSeleccionatInferior() {
    return(model.taula_yselec_inferior.hasValue(model.idRegistre_pintar));
  }
  public void setCanviGrafic() {
    vista.canviaRodes();
  }
  ////////////////////////////////////////////////////  

  public boolean isDialZero() {
    //Si és el registre zero (ja han passat 250 mil·lisegons), primer, cal pintar el dial que correspon al zero.
    return(model.idRegistre_pintar ==0);
  }
  public void inicia() {
    model.init();
  }
  public boolean pucAvansar() {
    return model.xDial <= model.xUltimDial;
  }

  //BASE DADES//////////////////////////////////////////////////// ///////////////////////////
  /**
   Classe per a encapsular les variables de valor fix.
   */
  private final class BD1 {
    //Nombre de frems per segon
    public final int fremsps = 32;
    //NOMS ARXIUS i Variables per a la llista desplegable.
    public  final String ARXIU_CAMA = "CAMA.CSV";
    public  final String ARXIU_CUIXA = "CUIXA.CSV";
    public  final String ARXIU_CINTURA = "CINTURA.CSV";
    public  final List llistaArxius = Arrays.asList(ARXIU_CINTURA, ARXIU_CUIXA, ARXIU_CAMA);
    public  final int Id_llista =17;
    public  final String NOM_LLISTA_ARXIUS ="llista_arxius";
    public  final int x_llista = BDPrincipal.AMPLE/2+315;
    public  final int y_llista = BDPrincipal.ALSADA/2+100;
    public  final int w_llista =210;
    public  final int h_llista = 100;
    public  final int h_item_llista =25;
    public  final int h_bar_llista = 20;
    //VARIABLES PER A GESTIONAR LES MIDES DE LA GRAFICA /////////////////////////  /////////////////////////  ///////////////////////// 
    //Per a definir els limits
    public  final int top_zona =30;
    public  final int bottom_zona =330;
    private  final int left = Principal.BDPrincipal.AMPLE/2-50;
    private  final int top = 0;
    public  final int w_caixa =600;
    public  final int h_caixa = 350;
    public  final int  w_roda = 7;        //Dimensions Rodes.
    public  final int  h_roda = 10;       //Dimensions Rodes.
    public  final int w_label=90;         //Dimensions etiquetes.
    public  final int y_dialTop = 22;
    public  final int y_dialBottom =328;
    public  final int x_fi_dials = w_caixa - 24;
    public  final int x_inici_dials = 96;
    public  final int w_dials = x_fi_dials-x_inici_dials;
    public  final int w_dial = 4;
    public  final int maxDials=  w_dials/w_dial;
    public  final int y_ZSuperiorTop = 32;
    public  final int y_ZSuperiorZero = 80;
    public  final int y_ZSuperiorBottom = 128;
    public  final int y_ZCentreTop = 132;
    public  final int y_ZCentreZero = 180;
    public  final int y_ZCentreBottom = 228;
    public  final int y_ZInferiorTop = 232;
    public  final int y_ZInferiorZero = 280;
    public  final int y_ZInferiorBottom = 328;
    //INDEX PER A GESTIONAR LES RODES /////////////////////////  /////////////////////////  ///////////////////////// 
    public final int  idr_roll_d = 0;
    public final int  idr_roll_e = 1;
    public final int  idr_pitch_r = 2;
    public final int  idr_pitch_p = 3; 
    public final int  idr_yaw = 4;
    public final int  idr_ox_p = 5;
    public final int  idr_ox_n = 6;
    public final int  idr_oy_p = 7;
    public final int  idr_oy_n = 8; 
    public final int  idr_oz_p = 9;
    public final int  idr_oz_n = 10;
    public final int  idr_fps = 11;
    public  final String  s_grau = "º";
    public  final String  s_mps = " m/s2";
    public  final String  s_fps = "fps";
    public final String s_zero = "0";
    //Taula per a indexar el nom de les rodes.
    public  final  String[] noms_r = {
      "Roll_rd", "Roll_re", 
      "Pitch_ra", "Pitch_rp", 
      "Yaw_r", 
      "ox_p", "ox_n", 
      "oy_p", "oy_n", 
      "oz_p", "oz_n", 
      "fps"};
    //INDEX PER A GESTIONAR LES ETIQUETES /////////////////////////  /////////////////////////  ///////////////////////// 
    public  final String s_pitch =new String(Character.toChars(unhex("03C6")));
    public  final String s_yaw =new String(Character.toChars(unhex("03C8")));
    public  final String s_roll = new String(Character.toChars(unhex("03A6")));
    public  final int idl_superior=0;
    public  final int idl_centre=2;
    public  final int idl_inferior=4;
    public  final int idl_hora=6;
    public  final  String[] noms_l ={ 
      "Roll_RY \n FRONTAL", "G_X   \n FRONTAL", 
      "Pitch_RX \n SAGITAL", "G_Y \n SAGITAL", 
      "Yaw_RZ \n HORITZONTAL", "G_Z \n HORITZONTAL", 
      "hora" };

    /**
     retorna el nom d'una etiqueta per a usar-lo
     com a contingut visual.
     */
    public String getValorVisual(int id) {
      return noms_l[id];
    }

    //Per a gestionar el boto play/pause /////////////////////////  /////////////////////////  ///////////////////////// 
    public  final String s_nom_play ="play"; 
    public  final String s_img_play ="play.jpg"; 
    public  final String s_img_play1 ="play1.jpg"; 
    public  final String s_img_pause ="pause.jpg"; 
    public  final String s_img_pause1 ="pause1.png";
  }
  //MODEL//////////////////////////////////////////////////// ///////////////////////////
  private class ModelGrafica {
    public boolean pause=false;
    public boolean isZonaGrafica=false;
    public boolean tenimArxiu = false;
    public boolean tenimPort = false;
    //Variable per a establir cada quants fotogrames volem pintar un registre.
    public int v_fotograma = BD.fremsps/4; //per defecte, cada 8 fotogrames.
    /////////////////////////////
    //Valors per defecte de les escales de dibuix
    public int roll_d_escala = 90; 
    public int roll_e_escala = -90; 
    public int pitch_a_escala = 180; 
    public int pitch_p_escala = -180;
    public int yaw_escala = 360;
    public int ox_p_escala = 15; 
    public int ox_n_escala = -15; 
    public int oy_p_escala = 15; 
    public int oy_n_escala = -15;
    public int oz_p_escala = 15;
    public int oz_n_escala = -15;
    ///////////////////////////////////////
    //per a la gràfica que mostra els valors de l'arxiu.
    public int idRegistre_pintar;
    public int idRowClicat ;
    public int idRowUltim ;
    public int idRowIniciMitja ;
    public int xUltim;
    public int xMouse;
    public int xDial;
    public int xUltimDial;
    //Per a desar els identificadors dels registres clicats.
    public  FloatList taula_yselec_superior = new FloatList(15);
    public  FloatList taula_yselec_centre = new FloatList(15);
    public  FloatList taula_yselec_inferior = new FloatList(15);
    /**
     Mètode cridat des de principal en clicar un punt mentre la grafica esta en pause.
     */
    public void seleccionarPunt(float yclicat) {
      //Si el punter està a sobre de la representació d'un registre i cliquem,
      //canviem les dades de l'últim registre  per les d'aquest.
      xUltim = xMouse;
      idRowUltim = idRowClicat;
      idRowIniciMitja = idRowClicat;
      //Si ja l'haviem clicat, l'esborrem, sino, el desem per a la llista.
      if (yclicat >= BD.y_ZSuperiorTop && yclicat <= BD.y_ZSuperiorBottom) {
        if (taula_yselec_superior.hasValue(idRowClicat))
          taula_yselec_superior.removeValue(idRowClicat);
        else
          taula_yselec_superior.append(idRowClicat);
      } else if (yclicat >= BD.y_ZCentreTop && yclicat <= BD.y_ZCentreBottom) {
        if (taula_yselec_centre.hasValue(idRowClicat))
          taula_yselec_centre.removeValue(idRowClicat);
        else
          taula_yselec_centre.append(idRowClicat);
      } else if (yclicat >= BD.y_ZInferiorTop && yclicat <= BD.y_ZInferiorBottom) {
        if (taula_yselec_inferior.hasValue(idRowClicat))
          taula_yselec_inferior.removeValue(idRowClicat);
        else 
        taula_yselec_inferior.append(idRowClicat);
      }
    }

    public void capturaDial() {
      // .. capturem l'id del registre i la posició del dial, 
      xUltim =xDial;
      idRowUltim = idRegistre_pintar;
    }

    /**
     mètode per a calcular el valors dels paràmetres per a dibuixar la gràfica.
     Establim els límits dels dials i el registre que toca pintar.
     */
    public void init() {
      //Coordenada x inici dels dials.
      xDial =BD.x_inici_dials ;
      //Coordenada x, final dels dials.
      xUltimDial = xDial + BD.w_dial*(mp.idRegistre % BD.maxDials);
      // Calculem el registre que toca pintar. 
      idRegistre_pintar = (mp.idRegistre/BD.maxDials)*BD.maxDials ;
      if (idRegistre_pintar < 0 ) idRegistre_pintar=0;
    }

    public boolean isPosicioDial() {
      //Calculem la diferencia entre el punter i la proximitat a un dial.
      int aux = mouseX-(BD.left + xDial);
      //Acceptem una desviació d'un pixel a cada costat
      return (aux >= -2 && aux < 2) ;
    }
    public String getValor(float v) {
      //Arrodoniex a 2 decimals.
      return str((int(round(v*10.0))/10.0));
    }
    public String getValorSuperior() {
      if (controladorPanel.isSeleccionat(controladorPanel.BD.GRAFICA_EULER))return getValor(mp.roll)+BD.s_grau;
      else if (controladorPanel.isSeleccionat(controladorPanel.BD.GRAFICA_GRAVETAT))return getValor(mp.ox)+BD.s_mps;
      return "";
    }
    public String getValorCentre() {
      if (controladorPanel.isSeleccionat(controladorPanel.BD.GRAFICA_EULER))return getValor(mp.pitch)+BD.s_grau;
      else if (controladorPanel.isSeleccionat(controladorPanel.BD.GRAFICA_GRAVETAT))return getValor(mp.oy)+BD.s_mps;
      return "";
    }
    public String getValorInferior() {
      if (controladorPanel.isSeleccionat(controladorPanel.BD.GRAFICA_EULER))return getValor(mp.yaw)+BD.s_grau;
      else if (controladorPanel.isSeleccionat(controladorPanel.BD.GRAFICA_GRAVETAT))return getValor(mp.oz)+BD.s_mps;
      return "";
    }
  }
  //VISTA LATERAL GRAFICA ZONA 3 //////////////////////////////////////////////////// ///////////////////////////
  private class VistaGraficaLateral {

    private ControladorRoda rd_roll, re_roll, ra_pitch, rp_pitch, r_yaw;
    private ControladorRoda rn_ox, rp_ox, rn_oy, rp_oy, rn_oz, rp_oz;
    private ControladorLabel l_temps, l_superior, l_centre, l_inferior;

    public VistaGraficaLateral() {
      creaLabels();
      creaRodesEuler();
      creaRodesGravetat();
    }

    private void creaRodesEuler() {
      int x = Principal.BDPrincipal.AMPLE/2 -45;
      int y = 35;
      rd_roll=  new ControladorRoda(x, y, BD.w_roda, BD.h_roda, BD.idr_roll_d, BD.noms_r[BD.idr_roll_d]);
      rd_roll.setRang(0, model.roll_d_escala, BD.s_grau);
      re_roll= new ControladorRoda(x, y+80, BD.w_roda, BD.h_roda, BD.idr_roll_e, BD.noms_r[BD.idr_roll_e]);
      re_roll.setRang(model.roll_e_escala, 0, BD.s_grau);
      ra_pitch= new ControladorRoda(x, y+100, BD.w_roda, BD.h_roda, BD.idr_pitch_r, BD.noms_r[BD.idr_pitch_r]);
      ra_pitch.setRang(0, model.pitch_a_escala, BD.s_grau);
      rp_pitch= new ControladorRoda(x, y+180, BD.w_roda, BD.h_roda, BD.idr_pitch_p, BD.noms_r[BD.idr_pitch_p]);
      rp_pitch.setRang(model.pitch_p_escala, 0, BD.s_grau);
      r_yaw= new ControladorRoda(x, y+200, BD.w_roda, BD.h_roda, BD.idr_yaw, BD.noms_r[BD.idr_yaw]);
      r_yaw.setRang(0, model.yaw_escala, BD.s_grau);
      model.roll_d_escala = 75;
      model.roll_e_escala = -75;
      model.pitch_a_escala = 75;
      model.pitch_p_escala = -75;
      setValorVisual(rd_roll, model.roll_d_escala);
      setValorVisual(re_roll, model.roll_e_escala);
      setValorVisual(ra_pitch, model.pitch_a_escala);
      setValorVisual(rp_pitch, model.pitch_p_escala);
    }
    private void creaRodesGravetat() {
      int x = Principal.BDPrincipal.AMPLE/2 -45;
      int y = 35;
      rp_ox=  new ControladorRoda(x, y, BD.w_roda, BD.h_roda, BD.idr_ox_p, BD.noms_r[BD.idr_ox_p]);
      rp_ox.setRang(0, model.ox_p_escala, BD.s_mps);
      rp_ox.setVisible(false);
      rn_ox=  new ControladorRoda(x, y+80, BD.w_roda, BD.h_roda, BD.idr_ox_n, BD.noms_r[BD.idr_ox_n]);
      rn_ox.setRang(0, model.ox_n_escala, BD.s_mps);
      rn_ox.setVisible(false);
      rp_oy=  new ControladorRoda(x, y+100, BD.w_roda, BD.h_roda, BD.idr_oy_p, BD.noms_r[BD.idr_oy_p]);
      rp_oy.setRang(0, model.oy_p_escala, BD.s_mps);
      rp_oy.setVisible(false);
      rn_oy=  new ControladorRoda(x, y+180, BD.w_roda, BD.h_roda, BD.idr_oy_n, BD.noms_r[BD.idr_oy_n]);
      rn_oy.setRang(0, model.oy_n_escala, BD.s_mps);
      rn_oy.setVisible(false);
      rp_oz=  new ControladorRoda(x, y+200, BD.w_roda, BD.h_roda, BD.idr_oz_p, BD.noms_r[BD.idr_oz_p]);
      rp_oz.setRang(0, model.oz_p_escala, BD.s_mps);
      rp_oz.setVisible(false);
      rn_oz=  new ControladorRoda(x, y+280, BD.w_roda, BD.h_roda, BD.idr_oz_n, BD.noms_r[BD.idr_oz_n]);
      rn_oz.setRang(0, model.oz_n_escala, BD.s_mps);
      rn_oz.setVisible(false);
    }

    private void canviaRodes() {
      if (controladorPanel.isSeleccionat(controladorPanel.BD.GRAFICA_EULER)) {
        rn_ox.setVisible(false);
        rn_oy.setVisible(false);
        rn_oz.setVisible(false);
        rp_ox.setVisible(false);
        rp_oy.setVisible(false);
        rp_oz.setVisible(false);
        rd_roll.setVisible(true);
        re_roll.setVisible(true);
        ra_pitch.setVisible(true);
        rp_pitch.setVisible(true);
        r_yaw.setVisible(true);
      } else if (controladorPanel.isSeleccionat(controladorPanel.BD.GRAFICA_GRAVETAT)) {
        rd_roll.setVisible(false);
        re_roll.setVisible(false);
        ra_pitch.setVisible(false);
        rp_pitch.setVisible(false);
        r_yaw.setVisible(false);
        rn_ox.setVisible(true);
        rp_ox.setVisible(true);
        rn_oy.setVisible(true);
        rp_oy.setVisible(true);
        //rn_oz.setVisible(true);
        rp_oz.setVisible(true);
      }
    }
    private void creaLabels() {
      l_temps = new ControladorLabel(BD.left, 12, 50, 15, BD.idl_hora, BD.noms_l[BD.idl_hora], "");
      l_superior = new ControladorLabel(BD.left, BD.y_ZSuperiorZero-25, BD.w_label, 50, BD.idl_superior, BD.getValorVisual(BD.idl_superior), 0xAAff0000);
      l_centre = new ControladorLabel(BD.left, BD.y_ZCentreZero-25, BD.w_label, 50, BD.idl_centre, BD.getValorVisual(BD.idl_centre), 0xAA00ff00);
      l_inferior= new ControladorLabel(BD.left, BD.y_ZInferiorZero-25, BD.w_label, 50, BD.idl_inferior, BD.getValorVisual(BD.idl_inferior), 0xAA0000ff);
    }


    public void setValorsLabels() {
      if (controladorPanel.isSeleccionat(controladorPanel.BD.GRAFICA_EULER)) {
        //Estic mostrant la gráfica dels valors EULER.
        l_superior.l.setStringValue(BD.getValorVisual(l_superior.l.getId())+"\n"+BD.s_roll +"="+model.getValor(mp.roll)+BD.s_grau);
        l_centre.l.setStringValue(BD.getValorVisual(l_centre.l.getId())+"\n"+BD.s_pitch +"="+model.getValor(mp.pitch)+BD.s_grau);
        l_inferior.l.setStringValue(BD.getValorVisual(l_inferior.l.getId())+"\n"+BD.s_yaw +"="+ model.getValor(mp.yaw)+BD.s_grau);
      } else if (controladorPanel.isSeleccionat(controladorPanel.BD.GRAFICA_GRAVETAT)) {
        l_superior.l.setStringValue(BD.getValorVisual(l_superior.l.getId()+1)+"\n= "+model.getValor(mp.ox/10.0)+ BD.s_mps);
        l_centre.l.setStringValue(BD.getValorVisual(l_centre.l.getId()+1)+"\n= "+model.getValor(mp.oy/10.0)+BD.s_mps);
        l_inferior.l.setStringValue(BD.getValorVisual(l_inferior.l.getId()+1)+"\n= "+model.getValor(mp.oz/10.0)+BD.s_mps);
      } else {
        l_superior.l.setStringValue(BD.getValorVisual(l_superior.l.getId()+1)+"\n= "+model.getValor(mp.oxA)+BD.s_mps);
        l_centre.l.setStringValue(BD.getValorVisual(l_centre.l.getId()+1)+"\n= "+model.getValor(mp.oyA)+BD.s_mps);
        l_inferior.l.setStringValue(BD.getValorVisual(l_inferior.l.getId()+1)+"\n= "+model.getValor(mp.ozA)+BD.s_mps);
      }
      l_temps.l.setStringValue(mp.hora);
    }
  }
}
