/** 
 Classe per a dibuixar la representació del vector Gravetat tal com s'ha definit a la memoria.
 */
public class ControladorGravetat {
  private ModelPrincipal mp = null;
  private VistaG vista= null;
  private BD1 BD = null;
  private PGraphics g = null;

  /**
   Mètode constructor, sol·licitat pel controladorPrincipal.
   */
  public ControladorGravetat(ModelPrincipal modelPrincipal) { 
    //Objectes per a implementar l'arquitectura MODEL_VISTA_CONTROLADOR
    BD = new BD1();
    mp =  modelPrincipal;
    vista = new VistaG();
  } 
  public void pintaEscena(PGraphics g ) {
    if (controladorPanel.isSeleccionat(controladorPanel.BD.GRAVETAT)) {
      this.g = g;
      g.textSize(10);
      vista.pintaVista();
    }
  }
  /**
   Mètode per a escriure les lletres que acompanyen al dibuix.
   */
  private void pintaText(float x, float y, float z, String s) {
    g.pushMatrix();
    //Anem on volem escriure.
    g.translate(x, y, z);
    //Fem la rotació inversa.
    mp.setMatriuTransposada(g);
    g.rotateX(PI/2);
    // Mirem de cara a l'observador.
    g.rotateY(radians(90 - controladorCamera.getTheta()));
    g.rotateX(radians(- controladorCamera.getPhi()));
    //Pintem el text
    g.text(s, -textWidth(s)/2, -2, 0);
    g.popMatrix();
  }
  //BASE DADES//////////////////////////////////////////////////// ///////////////////////////
  /**
   Classe per a encapsular les variables de valor fix.
   */
  private final class BD1 {
    //Mides.
    public  final float w_arc = 100f;//ample de l'arc.
    public  final float h_arc = 100f;//alçada de l'arc.
    public  final float w_arc_angle = 150f;//ample de l'arc.
    public  final float h_arc_angle = 150f;//alçada de l'arc.
    public  final int h_vector = 8;// hipotenusa del triangle que dibuixa la fletxa.
    public  final int c_vector = 2; //catet del triangle que dibuixa la fletxa.
    public  final int h_modul = 10;// hipotenusa del triangle que dibuixa la fletxa.
    public  final int c_modul = 3; //catet del triangle que dibuixa la fletxa.
    public  final int h_angle = 10;// hipotenusa del triangle que dibuixa la fletxa.
    public  final int c_angle = 4; //catet del triangle que dibuixa la fletxa.
    // llista Lletres.
    public  final String s_alfa = new String(Character.toChars(unhex("03B1")));
    public  final String s_beta = new String(Character.toChars(unhex("03B2")));
    public  final String s_gamma = new String(Character.toChars(unhex("03B3")));
    public  final String  s_theta = new String(Character.toChars(unhex("03B8"))); 
    public  final String  s_phi = new String(Character.toChars(unhex("03C6")));
    public final int id_A = 0; 
    public final int id_P = 1;
    public final int id_O = 2;
    public final int id_x = 3;
    public final int id_y = 4;
    public final int id_z = 5;
    public final int id_i = 6;
    public final int id_j = 7;
    public final int id_k = 8;
    private final StringList noms =new StringList("A", "P", "O", "x", "y", "z", "i", "j", "k");
    public String getText(int id) {
      return noms.get(id);
    }
  }


  //VISTA//////////////////////////////////////////////////// ///////////////////////////
  private class VistaG {

    public void pintaVista() {
      pintaCaixa();
      pintaModul();
      pintaProjeccioModul();
      pintaVector_i();
      pintaVector_j(); 
      pintaVector_k();
      pintaAnglePhi();
      pintaAngleTheta();
    }

    private void pintaCaixa() {
      //Dibuixem la caixa.
      g.stroke(0, 0, 0); 
      g.strokeWeight(1);
      g.line(0, 0, 0, mp.ox, 0, 0);
      g.line(0, 0, 0, 0, mp.oy, 0);
      g.line(0, 0, 0, 0, 0, mp.oz);
      g.line(mp.ox, 0, mp.oz, 0, 0, mp.oz);
      g.line(mp.ox, 0, mp.oz, mp.ox, 0, 0);
      g.line(mp.ox, mp.oy, 0, mp.ox, 0, 0);
      g.line(mp.ox, mp.oy, 0, 0, mp.oy, 0);
      g.line(0, mp.oy, mp.oz, 0, mp.oy, 0);
      g.line(0, mp.oy, mp.oz, 0, 0, mp.oz);
      g.line(mp.ox, mp.oy, mp.oz, mp.ox, 0, mp.oz);
      g.line(mp.ox, mp.oy, mp.oz, mp.ox, mp.oy, 0);
      g.line(mp.ox, mp.oy, mp.oz, 0, mp.oy, mp.oz);
      //Pintem les lletres.
      g.fill(0, 0, 0);
      pintaText(mp.ox, (abs(mp.oy)/(mp.oy+0.1))*4, (abs(mp.oz)/(mp.oz+0.1))*4, BD.getText(BD.id_x));
      pintaText((abs(mp.ox)/(mp.ox+0.1))*4, mp.oy, (abs(mp.oz)/(mp.oz+0.1))*4, BD.getText(BD.id_y));
      pintaText((abs(mp.ox)/(mp.ox+0.1))*4, (abs(mp.oy)/(mp.oy+0.1))*4, mp.oz, BD.getText(BD.id_z));
    }  

    private void pintaModul() {
      g.strokeWeight(3);
      g.stroke(255, 0, 255);
      g.fill(255, 0, 255);
      g.line(0, 0, 0, mp.ox, mp.oy, mp.oz);
      g.pushMatrix();  // // // // // // // // // // // // // // //  pushMatrix()
      //Anem on volem la punta de fletxa.
      g.translate( mp.ox, mp.oy, mp.oz);
      //recuperem la posició original
      mp.setMatriuTransposada(g);
      g.rotateX(-PI/2);
      // Mirem de cara a l'observador.
      g.rotateY(radians(controladorCamera.getTheta()));
      //Pintem la punta de fletxa.
      g.beginShape(TRIANGLE);
      g.vertex(0, 0, 0);
      g.vertex(0, BD.h_modul, -BD.c_modul); 
      g.vertex(0, BD.h_modul, BD.c_modul);
      g.endShape();
      g.popMatrix(); //// // // // // // // // // // // // // //  popMatrix()
      //Pintem el text
      pintaText(mp.ox, mp.oy, mp.oz+15, BD.getText(BD.id_A));
    } 

    private void pintaProjeccioModul() {
      g.stroke(0, 0, 255);
      g.line(0, 0, 0, mp.ox, mp.oy, 0); 
      //Pintem un punt
      g.pushMatrix();  // // // // // // // // // // // // // // //  pushMatrix()
      g.translate( mp.ox, mp.oy, 0);
      g.sphere(1);
      g.popMatrix(); // // // // // // // // // // // // // // //  popMatrix()
      //Pintem el text
      g.stroke(0, 0, 0);
      pintaText(  mp.ox/5, mp.oy/5, -(abs( mp.oz)/( mp.oz+0.1))*2, BD.getText(BD.id_O));
      pintaText( mp.ox+2, mp.oy+2, -(abs( mp.oz)/( mp.oz+0.1))*2, BD.getText(BD.id_P));
    }

    private void pintaVector_j() {
      g.stroke(255, 0, 255);
      g.strokeWeight(1);
      //Segment per unir els punts OX amb OA
      g.line(0, mp.oy, 0, mp.ox, mp.oy, mp.oz);
      //Segment per a representar el vector unitari i la lletra.
      g.strokeWeight(2);
      g.line(0, 0, 0, 0, mp.oy/2, 0);
      g.fill(0, 0, 0); 
      pintaText(0, mp.oy/2, 0, BD.getText(BD.id_j));
      g.pushMatrix(); // // // // // // // // // // // // // // //  pushMatrix()
      //Pintem la fletxa del segment.
      g.fill(255, 0, 255); 
      //ens traslladem al final de l'arc i pintem la punta de fletxa.
      g.translate(0, mp.oy/2, 0);
      g.beginShape(TRIANGLE);
      if (degrees( mp.aj)< 0) g.vertex(0, BD.h_vector, 0);
      else g.vertex(0, -BD.h_vector, 0);
      g.vertex(-BD.c_vector, 0, 0); 
      g.vertex(BD.c_vector, 0, 0);
      g.endShape();
      g.popMatrix();// // // // // // // // // // // // // // //  popMatrix()
      g.pushMatrix();  // // // // // // // // // // // // // //  pushMatrix()
      //Rotem fins trobar el vector.
      if ( mp.theta_g >=180) g.rotateY(-atan( mp.oz/ mp.ox));
      else if ( mp.theta_g < 180) g.rotateY(PI-atan( mp.oz/ mp.ox));
      //Pintem l'arc.
      g.fill(255, 255, 0, 80);
      g.strokeWeight(1);
      if ( mp.oy <0) {
        g.arc(0.0, 0.0, -BD.w_arc, -BD.h_arc, PI/2, PI/2+ mp.aj, PIE);
        //Anem a la posició on escriurem la lletra de l'angle.
        g.translate(40 * sin( mp.aj/2), -40*cos( mp.aj/2), 0);
      } else if ( mp.oy > 0) {
        g.arc(0.0, 0.0, BD.w_arc, BD.h_arc, PI/2+ mp.aj, PI/2, PIE);
        g.translate(-40 * sin( mp.aj/2), 40*cos( mp.aj/2), 0);
      }
      //Recuperem el sistema de referencia anterior.
      if ( mp.theta_g >=180) g.rotateY(atan( mp.oz/ mp.ox));
      else  g.rotateY(PI+atan( mp.oz/ mp.ox));
      //pintem la lletra de l'angle.
      g.fill(0, 0, 0); 
      pintaText( (abs( mp.ox)/( mp.ox+0.1))*4, 0, (abs( mp.oz)/( mp.oz+0.1))*4, BD.s_beta);
      g.popMatrix();// // // // // // // // // // // // // // //  popMatrix()
    }

    private void pintaVector_i() {
      g.stroke(255, 0, 255);
      g.strokeWeight(1);
      //Segment per unir els punts  mp.ox amb OA
      g.line( mp.ox, 0, 0, mp.ox, mp.oy, mp.oz);
      //Segment per a representar el vector unitari i la lletra.
      g.strokeWeight(2);
      g.line(0, 0, 0, mp.ox/2, 0, 0);
      g.fill(0, 0, 0); 
      pintaText( mp.ox/2, 0, 0, BD.getText(BD.id_i));
      g.pushMatrix(); // // // // // // // // // // // // // // //  pushMatrix()
      //Pintem la fletxa del segment.
      g.fill(255, 0, 255); 
      g.translate( mp.ox/2, 0, 0);
      g.beginShape(TRIANGLE);
      if (degrees( mp.ai)< 0) g.vertex(-BD.h_vector, 0, 0);
      else   g.vertex(BD.h_vector, 0, 0);
      g.vertex(0, -BD.c_vector, 0); 
      g.vertex(0, BD.c_vector, 0);
      g.endShape();
      g.popMatrix();// // // // // // // // // // // // // // //  popMatrix()
      g.pushMatrix();  // // // // // // // // // // // // // //  pushMatrix()
      //Rotem fins trovar el vector.
      if ( mp.theta_g >= 90 &&  mp.theta_g <270) g.rotateX(atan( mp.oz/ mp.oy));
      else   g.rotateX(atan( mp.oz/ mp.oy)+PI);
      //Pintem l'arc.
      g.fill(255, 255, 0, 80);
      g.strokeWeight(1);
      if ( mp.ox <0) {
        g.arc(0.0, 0.0, -BD.w_arc, -BD.h_arc, mp.ai, 0, PIE);
        //Anem a la posició on escriurem la lletra de l'angle.
        g.translate(-40 * cos( mp.ai/2), -40*sin( mp.ai/2), 0);
      } else if ( mp.ox > 0) {
        g.arc(0.0, 0.0, BD.w_arc, BD.h_arc, 0, mp.ai, PIE);
        g.translate(40 * cos( mp.ai/2), 40*sin( mp.ai/2), 0);
      }
      //Recuperem el sistema de referencia anterior.
      if ( mp.theta_g >= 90 &&  mp.theta_g <270) g.rotateX(-atan( mp.oz/ mp.oy));
      else   g.rotateX(-atan( mp.oz/ mp.oy)+PI);
      //pintem la lletra de l'angle.
      g.fill(0, 0, 0); 
      pintaText(0, (abs( mp.oy)/( mp.oy+0.1))*-4, (abs( mp.oz)/( mp.oz+0.1))*4, BD.s_alfa);
      g.popMatrix();// // // // // // // // // // // // // // //  popMatrix()
    }

    private void pintaVector_k() {
      g.stroke(255, 0, 255);
      g.strokeWeight(1);
      //Segment per unir els punts  mp.ox amb OA
      g.line(0, 0, mp.oz, mp.ox, mp.oy, mp.oz);
      //Segment per a representar el vector unitari i la lletra.
      g.strokeWeight(2);
      g.line(0, 0, 0, 0, 0, mp.oz/2);
      g.fill(0, 0, 0); 
      pintaText(0, 0, mp.oz/2, BD.getText(BD.id_k));
      g.pushMatrix();  // // // // // // // // // // // // // // //  pushMatrix()
      g.rotateZ(radians(- mp.theta_g)-PI/2);
      g.pushMatrix(); // // // // // // // // // // // // // // //  pushMatrix()
      //Pintem la fletxa del segment.
      g.fill(255, 0, 255); 
      g.translate(0, 0, mp.oz/2);
      g.beginShape(TRIANGLE);
      if (degrees( mp.ak)< 0) g.vertex(0, 0, -BD.h_vector);
      else  g.vertex(0, 0, BD.h_vector);
      g.vertex(0, -BD.c_vector, 0); 
      g.vertex(0, BD.c_vector, 0);
      g.endShape();
      g.popMatrix();// // // // // // // // // // // // // // //  popMatrix()
      //Rotem fins trobar el vector.
      //Rotem per a pintar l'arc de phi
      if ( mp.oz<0) g.rotateX(-PI/2);
      else g.rotateX(PI/2);
      if ( mp.phi_g<0) g.rotateZ(-radians( mp.phi_g));
      else  g.rotateZ(radians( mp.phi_g));
      //Pintem l'arc.
      g.fill(255, 255, 0, 80);
      g.strokeWeight(1);
      if ( mp.oz<0) g.arc(0.0, 0.0, BD.w_arc, -BD.h_arc, mp.ak, 0, PIE);
      else  g.arc(0.0, 0.0, BD.w_arc, BD.h_arc, 0, mp.ak, PIE);
      //Anem a la posició on pintem la lletra de l'angle.
      g.translate(35 * cos( mp.ak), 35*sin( mp.ak), 0);
      //Recuperem el sistema de referencia anterior.
      if ( mp.phi_g<0)g.rotateZ(radians( mp.phi_g));        
      else g.rotateZ(radians(- mp.phi_g));
      if ( mp.oz<0) g.rotateX(PI/2);
      else g.rotateX(-PI/2);
      g.rotateZ(radians( mp.theta_g)+PI/2);
      g.fill(0, 0, 0);   
      pintaText( (abs( mp.ox)/( mp.ox+0.1))*8, (abs( mp.oy)/( mp.oy+0.1))*4, 0, BD.s_gamma);
      g.popMatrix();// // // // // // // // // // // // // // //  popMatrix()
    }

    private void pintaAngleTheta() {
      g.stroke(0, 0, 255);
      g. pushMatrix();// // // // // // // // // // // // // // // pushMatrix() 
      g.fill(0, 0, 255); //Pintem el simbol de Theta.
      pintaText(BD.w_arc_angle/2 * cos(radians( mp.theta_g+BD.w_arc_angle/2)), -BD.w_arc_angle/2 * sin(radians( mp.theta_g+80)), 0, BD.s_theta);
      g.fill(0, 0, 255, 40); 
      g.rotateZ(-PI/2); //Rotem Z per tal de pintar l'angle tal com l'entrega la IMU.
      g.arc(0, 0, BD.w_arc_angle, -BD.h_arc_angle, 0, radians( mp.theta_g), PIE);
      g. translate(BD.w_arc_angle/2 *  mp.cy, -BD.w_arc_angle/2 * mp.sy, 0);      //Anem a la punta.
      g.rotateZ(radians(- mp.theta_g));     //Centrem els eixos.    
      g.fill(0, 0, 255); //Pintem el simbol de Theta.
      g.beginShape(TRIANGLE);
      g.vertex(0, BD.h_angle, 0);
      g.vertex(BD.c_angle, BD.h_angle, 0); 
      g.vertex(0, 0, 0);
      g.endShape();
      g.popMatrix();// // // // // // // // // // // // // // //  popMatrix()
    }

    private void pintaAnglePhi() {
      //Posicio dels eixos normal.
      g.pushMatrix();  // // // // // // // // // // // // // // //  pushMatrix()
      //Rotem fins theta
      g.rotateZ(-PI/2);
      g.rotateZ(-radians( mp.theta_g));
      g.stroke (0, 150, 0); 
      g.fill(0, 150, 0); 
      g.pushMatrix();
      if ( mp.phi_g<0) g.translate(85 * cos(radians( mp.phi_g/2)), 0, -85 * sin(radians( mp.phi_g/2)));
      else g.translate(85 * cos(radians( mp.phi_g/2)), 0, 85 * sin(radians( mp.phi_g/2)));
      g.rotateZ(PI/2);
      pintaText(0, 0, 0, BD.s_phi);
      g.popMatrix();
      //Rotem per a pintar l'arc de phi
      if ( mp.oz<0) g.rotateX(-PI/2);
      else g.rotateX(PI/2);
      if ( mp.phi_g<0) {
        //Pintem l'arc.
        g.fill(0, 255, 0, 40); 
        g.stroke (0, 150, 0); 
        g.arc(0.0, 0.0, BD.w_arc_angle, -BD.h_arc_angle, radians( mp.phi_g), 0, PIE);
        g.fill(0, 150, 0);
        g.translate(BD.w_arc_angle/2 *  mp.cp, -BD.w_arc_angle/2 *  mp.sp, 0);
        g.rotateZ(-radians( mp.phi_g-90));
        g.beginShape(TRIANGLE);
        g.vertex(0, 0, 0);
        g.vertex(-BD.h_angle, 0, 0);
        g.vertex(-BD.h_angle, -BD.c_angle, 0); 
        g.endShape();
      } else {  
        g.fill(0, 255, 0, 40); 
        g.stroke (0, 150, 0); 
        g.arc(0.0, 0.0, BD.w_arc_angle, BD.h_arc_angle, 0, radians( mp.phi_g), PIE);
        g.fill(0, 150, 0);
        g.translate(BD.w_arc_angle/2 *  mp.cp, BD.w_arc_angle/2 *  mp.sp, 0);
        g.rotateZ(radians( mp.phi_g-90));
        g.beginShape(TRIANGLE);
        g.vertex(BD.h_angle, 0, 0);
        g.vertex(BD.h_angle, BD.c_angle, 0); 
        g.vertex(0, 0, 0);
        g.endShape();
      }
      g.popMatrix();// // // // // // // // // // // // // // //  popMatrix()
    }
  }
}
