/**
 Classe per a impelmentar el Model per a gestionar les dades que obtenim dels registres.
 */
public final class ModelPrincipal {
  // Variables per a la matriu de orientació.
  private float  r11, r12, r13, r21, r22, r23, r31, r32, r33;  
  //Variables per a la gráfica.
  public Table table;
  public int  idRegistre = -1;
  public TableRow rowActual ;
  //Variables angles Euler
  public float roll, pitch, yaw;
  // Variables per a la gravetat
  public float phi_g, theta_g;
  public  float ai, aj, ak;
  public float ox, oy, oz, op, modul;
  private float  sp, cp, sy, cy;
  // Variables per a l'acceleració
  private float oxA, oyA, ozA;
  // Variables per al control del calibrat.
  public int  sys, gyro, accel, mag ;
  //Variable per a mostrar l'hora.
  public String hora;


  public ModelPrincipal() {
  }

  /**
   Mètode per a iniciar la taula per a desar els registres que anirem pintant 
   quan llegim un arxiu.
   */
  public void iniciaTaula(String nomArxiu) {
    table = loadTable(nomArxiu, "header");
  }

  /**
   Mètode per a iniciar la taula per a desar els registres que anirem pintant 
   en temps real.
   */
  public void iniciaTaula() {
    //Instanciem l'objecte Table amb el que treballarem.
    table = new Table();
    table.addColumn("0", Table.STRING); 
    table.addColumn("1", Table.FLOAT); 
    table.addColumn("2", Table.FLOAT); 
    table.addColumn("3", Table.FLOAT);
    table.addColumn("4", Table.FLOAT); 
    table.addColumn("5", Table.FLOAT); 
    table.addColumn("6", Table.FLOAT);
    table.addColumn("7", Table.FLOAT); 
    table.addColumn("8", Table.FLOAT); 
    table.addColumn("9", Table.FLOAT);
    table.addColumn("10", Table.INT);
    table.addColumn("11", Table.INT); 
    table.addColumn("12", Table.INT); 
    table.addColumn("13", Table.INT);
    // Per cada registre que podem representar l'ample de la gràfica, afegim un Row
    for (int i =0; i< controladorGrafica.BD.maxDials; i++) 
      table.addRow();
  }

  /**
   Mètode per a desar un registre llegit pel port Serial a la taula de treball.
   En temps real.
   */
  public void setRegistre(String[] valors) {
    // incrementem el comptador.
    idRegistre++;
    //Quan arribem al final de la taula, tornem a desar els valors al principi de la taula.
    if (idRegistre == table.lastRowIndex()) idRegistre =0;

    //Hora a la que s'ha fet el registre.
    table.setString(idRegistre, 0, valors[0]);
    // Angle YAW
    table.setFloat(idRegistre, 1, float(valors[1]));
    //Angle ROLL
    table.setFloat(idRegistre, 2, float(valors[2]));
    //Angle PITCH
    table.setFloat(idRegistre, 3, float(valors[3]));
    // Gravetat X
    table.setFloat(idRegistre, 4, float(valors[4]));
    // Gravetat Y
    table.setFloat(idRegistre, 5, float(valors[5]));
    // Gravetat Z
    table.setFloat(idRegistre, 6, float(valors[6]));
    // Acceleració X
    table.setFloat(idRegistre, 7, float(valors[7]));
    // Acceleració Y
    table.setFloat(idRegistre, 8, float(valors[8]));
    // Acceleració Z
    table.setFloat(idRegistre, 9, float(valors[9]));
    //  Calibrat Gyroscopi
    table.setInt(idRegistre, 10, int(valors[10]));
    //  Calibrat Acceleracio
    table.setInt(idRegistre, 11, int(valors[11]));
    // Calibrat Magnetometre
    table.setInt(idRegistre, 12, int(valors[12]));
    // Calibrat Sistema
    table.setInt(idRegistre, 13, int(valors[13]));
  }

  public void setMatriuTransformacio(PGraphics g) {
    //apliquem la matriu de rotació.
    g.applyMatrix(
      r11, r12, r13, 0, 
      r21, r22, r23, 0, 
      r31, r32, r33, 0, 
      0, 0, 0, 1);
  }
  public void setMatriuTransposada(PGraphics g) {
    //apliquem la matriu de rotació inversa.
    g.applyMatrix(
      r11, r21, r31, 0, 
      r12, r22, r32, 0, 
      r13, r23, r33, 0, 
      0, 0, 0, 1);
  }
  /**
   Mètode per a llegir un registre de la taula de treball i
   carregar-lo en les variables.
   */
  public void carregaRegistre(int idRegistre) {
    //Controlem el desbordament dels valors.
    if (idRegistre >= table.lastRowIndex() || idRegistre < 0 ) 
      idRegistre=0;
    //Obtenim el registre. i carrgeum les variables
    TableRow row = table.getRow(idRegistre);
    // hora
    hora = formatHora(row.getString(0));
    //Orientació
    yaw =  row.getFloat(1);
    roll=  row.getFloat(2);
    pitch=  row.getFloat(3);
    calculaMatriuTransformacio();
    //Gravetat, mutipliquem per 10 per a fer el dibuix.
    ox= row.getFloat(4)*10f;
    oy= row.getFloat(5)*10f;
    oz =row.getFloat(6)*10f;
    calculaVectorGravetat();
    //Acceleració
    oxA= row.getFloat(7);
    oyA= row.getFloat(8);
    ozA =row.getFloat(9);
    //Calibrat.
    gyro =row.getInt(10);
    accel= row.getInt(11);
    mag =row.getInt(12);
    sys = row.getInt(13);
  }

  /**
   Mètode per a incrementar el comptador de l’últim registre que s'ha de pintar.
   Explicació: en el nostre cas, la pantalla es refresca 32 cops per segon. Per tal
   de crear l’animació de a gràfica, pintem un nou registre cada 8 fotogrames (valor per defecte).
   */
  public void avansaIDRegistre() {
    // Si no estem en pause, incrementem el comptador per tal d'accedir a la següent dada.
    if (!controladorGrafica.isPause())  idRegistre++;
    //Controlem el desbordament dels valors.
    if (idRegistre >=  table.lastRowIndex() || idRegistre < 0 ) 
      idRegistre=0;
  }

  private void calculaMatriuTransformacio() {
    float sp, cp, sr, cr, sy, cy;
    sy = sin(radians(yaw));
    cy = cos(radians(yaw));  
    sr = sin(radians(roll));
    cr = cos(radians(roll));
    sp = sin(radians(pitch));
    cp = cos(radians(pitch));
    r11 = cy*cr;
    r12 = sy*cp + cy*sp*sr;
    r13 = sy*sp - cy*sr*cp;
    r21 = -cr*sy;
    r22 = cy*cp - sr*sp*sy;
    r23 = cy*sp + sy*sr*cp;
    r31 = sr;
    r32 = -cr*sp;
    r33 = cp*cr;
  }

  private void calculaVectorGravetat() {
    //Dividim per 10 per a eliminar l'escala de dibuix que hem possat al principi.
    float x = ox/10f;
    float y = oy/10f;
    float z = oz/10f;
    //Mòdul del vector.
    modul = sqrt(x*x+y*y+z*z);
    //projecció del modul en el pla horitzontal.
    op = sqrt(x*x+y*y);
    //Calculem el valor de l'angle phi, el sinus i cosinus
    if (x> 0 ) phi_g=degrees(atan(abs(z)/op));
    else phi_g=-degrees(atan(abs(z)/op));
    sp = sin(radians(phi_g));
    cp = cos(radians(phi_g));
    //Calculem el valor de l'angle theta, el sinus i cosinus
    //Volem que theta comenci a comptar des de l'eix Sagital- davant
    if (x<0)theta_g = 90-degrees(atan(y/x));
    else theta_g = 270-degrees(atan(y/x));
    sy = sin(radians(theta_g));
    cy = cos(radians(theta_g));
    //Calculem l'angle entre OA i l'eix X
    //Calculem el catet XA, la relació entre catets, i obtenim l'angle.
    ai = atan(sqrt(modul*modul - x*x)/x);
    //Calculem l'angle entre OA i l'eix Y
    aj = atan(-sqrt(modul*modul - y*y)/y);
    //Calculem l'angle entre OA i l'eix Z
    ak = atan(sqrt(modul*modul - z*z)/z);
  }

  private String formatHora(String v) {
    //Preparem el format del temps, volem > hh:mm:ss i podem rebre, p.e:  hh:m:s
    String hora[]=null;
    String t ="00:00:00";
    try {
      if (v!=null) {
        hora = split(trim(v), ':');
        for (String s : hora) {
          if (s.length() ==1) s ='0'+s;
          if (t.equals("00:00:00")) t = s;
          else t = t+":"+s;
        }
      }
    } 
    //Capturem les excepcions
    catch (java.lang.ArrayIndexOutOfBoundsException e) {
      Log.printInfo(new Exception(), e.toString());
    }  
    catch (java.lang.NullPointerException e) {
      Log.printInfo(new Exception(), e.toString());
    }
    return t;
  }

  /**
   Mètode per a calcular la mitja entre el valor clicat i l'ùltim valor representat.
   Calcula la mitja per separat de valors negatius i positius.
   */
  public float getYMitja(int idZona, int idInici) {
    float primer = table.getRow(idInici).getFloat(idZona);
    float mitja = primer;
    float n =1;
    if (primer<0.0) {//valors negatius
      for (int i =idInici+1; i <= idRegistre; i++) {
        float v = table.getRow(i).getFloat(idZona);
        if (v < 0.0) {
          mitja = mitja +v;
          n++;
        }
      }
    } else {//valors positius.
      for (int i =idInici+1; i <= idRegistre; i++) {
        float v = table.getRow(i).getFloat(idZona);
        if (v>0.0) {
          mitja = mitja +v;
          n++;
        }
      }
    }
    println("Inici__fi: "+idInici+ ";"+ idRegistre);
    println("trobats: "+n + " mitja: "+ (mitja/n));
    return (mitja/n);
  }
}
