 //<>//

/**
 Classe per a pintar la gràfica quan estem llegint les dades d'un arxiu.
 La gràfica té tres quatre zones diferenciades:
 - la superior, per a pintar els valors positius i negatius de l'angle Roll o el X del vector gravetat.
 - la central, per a pintar els valors positius i negatius de l'angle Pitch o el Y del vector gravetat.
 - la inferior, per a pintar els valors l'angle Yaw o el Z del vector gravetat.
 - L'esquerra, que conté per a cadascuna de les anteriors: 
 -- Una etiqueta interactiva que informa dels valors relacionats amb l'angle o la coordenada.
 -- Una roda per a poder modificar l'escala de dibuix.
 La part esquerra  equipada amb un boto per aturar l'animació i una roda per accelerar la velocitat de dibuix de l'últim registre.
 Els punts de l’últim dial, tindràn un  diàmetre de valor 4, color interior Magenta i negre exterior.
 Els punts anteriors a l’últim, tindran un  diàmetre de valor 1 i color negre.
 //CAS D'US: L'usuari clica el boto pause:
 //REQUISIT_1: l'animació de la gràfica s'atura.
 ////SUBCAS: L'usuari desplaça el punter pel damunt dels dials:
 //////REQUISIT_1: el punt que apunta el dial es pintarà com si fos l'últim i totes les escenes també.
 //////REQUISIT_2: el que és últim quedarà pintat en magenta.
 ////////SUBCAS: L'usuari desplaça el punter fora de la zona de dials:
 //////////REQUISIT_1: totes les escenes es pinten amb el valor de l'ùltim punt.
 ////////SUBCAS: L'usuari clica un dial:
 //////////REQUISIT_1: el punt clicat passa a ser l'últim.
 */

public class VistaGraficaSD { 
  public PGraphics g;
  private  Button b_playPause;
  private ControladorRoda r_frame=null;
  private ScrollableList ls = null;
  public ControladorGrafica.ModelGrafica model;
  public ControladorGrafica.BD1 BD;
  public ControladorGrafica controlador;

  VistaGraficaSD (ControladorGrafica controlador) {
    //Objectes per a implementar l'arquitectura MODEL_VISTA_CONTROLADOR
    this.model = controlador.getModel();
    this.BD = controlador.getBD();
    this.controlador=controlador;
    //Objecte on pintem la gràfica.
    g =createGraphics(BD.w_caixa, BD.h_caixa);
  }

  private void init() {
    //Instanciem els botons.
    //Això ho tindríem que fer a ControladorGrafica, però ja té masses línies de codi.
    creaBotons();
  }

  /**
   Metode per a pintar els eixòs que separen les diferents zones.
   */
  public void pintaEixos() { 
    g.background(255, 255, 255);
    g.stroke(0, 0, 0);
    g.fill (0, 0, 0);
    g.strokeWeight(1);
    g.textSize(12);
    String s = BD.s_zero + BD.s_grau;

    //Recta Zona Superior, valor superior 90º
    g.line(0, BD.y_ZSuperiorTop, BD.x_fi_dials, BD.y_ZSuperiorTop);
    //Pintem l'eix que representa el valor zero (0º).
    g.line(BD.w_label, BD.y_ZSuperiorZero, BD.x_fi_dials, BD.y_ZSuperiorZero);
    g.text(s, BD.x_fi_dials - textWidth(s), BD.y_ZSuperiorZero-3);
    //Recta Zona Superior, valor inferior -90º
    g.line(0, BD.y_ZSuperiorBottom, BD.x_fi_dials, BD.y_ZSuperiorBottom);

    //Recta Zona Centre, valor superior 180º
    g.line(0, BD.y_ZCentreTop, BD.x_fi_dials, BD.y_ZCentreTop);
    //Pintem l'eix que representa el valor zero (0º).
    g.line(BD.w_label, BD.y_ZCentreZero, BD.x_fi_dials, BD.y_ZCentreZero);
    g.text(s, BD.x_fi_dials - textWidth(s), BD.y_ZCentreZero-3);
    //Recta Zona Centre, valor inferior -180º
    g.line(0, BD.y_ZCentreBottom, BD.x_fi_dials, BD.y_ZCentreBottom);

    //Recta Zona Inferior, valor superior 360º
    g.line(0, BD.y_ZInferiorTop, BD.x_fi_dials, BD.y_ZInferiorTop);
    //Pintem l'eix que representa el valor zero (0º).
    g.line(0, BD.y_ZInferiorBottom, BD.x_fi_dials, BD.y_ZInferiorBottom); 
    g.text(s, BD.x_fi_dials - textWidth(s), BD.y_ZInferiorBottom-3);
    //Text per als nombres dels dials.
    g.textSize(8);
  }

  /**
   Mètode per  a pintar la gràfica.
   */
  public void pintaEscena() { 
    g.beginDraw(); 
    pintaEixos();
    controlador.inicia();
    if (controlador.isDialZero()) pintaDialZero();
    //Pintem tots els registres fins l'actual
    while (controlador.pucAvansar()) {
      //Demanem un registre per a pintar.
      controlador.getRegistre();
      //Pintem el dial.
      pintaDial();
      if (model.pause) pintaRegistrePause();
      else pintaRegistre();
      controlador.avansa();
    }
    //Si estem en pause, quan sortim de la zona de la gráfica
    //pinta l'últim  registre.
    if (model.pause  && !model.isZonaGrafica)
      pintaUltim();
    g.endDraw(); 
    image(g, BD.left, BD.top);
  }

  protected void pintaDial() {
    // cada 4 registres, pintem el dial encapçalat  pel segon de temps que representa i d'un altre color.
    if ((model.idRegistre_pintar+1) % 4 ==0) {// el dial 1 correspon al quart registre (0,1,2,3)
      g.fill(0, 0, 0);
      g.stroke (0, 0, 0);
      String s =str((model.idRegistre_pintar+1)/4 );
      if (s.length()==1) g.text(s, model.xDial-(textWidth(s)/2.0-3), BD.y_ZSuperiorTop-15);
      else if (s.length()==2) g.text(s, model.xDial-(textWidth(s)/2.0 -6), BD.y_ZSuperiorTop-15);
      else if (s.length()==3) g.text(s, model.xDial-(textWidth(s)/2.0 -7), BD.y_ZSuperiorTop-15);
      g.stroke (255, 0, 0);
      g.line(model.xDial, BD.y_dialTop, model.xDial, BD.y_dialBottom);
    } else {
      // Si no, pintem l'eix vertical del mil·lisegon.
      g.stroke (0, 0, 0, 60);
      g.line(model.xDial, BD.y_dialTop+5, model.xDial, BD.y_dialBottom);
    }
  }

  /**
   Mètode per a pintar le dial zero al principi de la visualització dels registres.
   */
  public void pintaDialZero() {
    g.text("0", model.xDial-BD.w_dial-(textWidth("0")/2.0 -2), BD.y_dialTop-5);
    g.stroke (255, 0, 0);
    g.line(model.xDial-BD.w_dial, BD.y_dialTop, model.xDial-BD.w_dial, BD.y_dialBottom);
  }

  /**
   Mètode per actualitzar la vista amb els valors del registre que toca.
   Estem visualitzant les dades d'un arxiu.
   */
  public void pintaRegistre() {
    //l'últim registre el pintem diferent i actualitzem tots els valors.
    if (controlador.isUltimRegistre()) {
      g.stroke (0, 0, 0);
      g.fill(255, 0, 255);
      pintaPunt(4, 4);
      //Actulitzem el valor de les etiquetes.
      controlador.resetValorsLabelsLateral();
      //Demanem actualitzar el panell de dades.
      controladorPanel.resetlabelsGravetat();
      controladorPanel.resetValorsCalibrat();
    } //sinó, el pintem normal.
    else {
      g.stroke (0, 0, 0);
      g.fill(0, 0, 0);
      pintaPunt(1, 1);
    }
  }

  /**
   Mètode per actualitzar la vista amb els valors del registre que toca.
   Estem visualitzant les dades d'un arxiu i estem en pause.
   */
  private void pintaRegistrePause() {
    //Si estem en pause i movem el mouse..., 
    if ( controlador.isRegistrePause()) {
      //pintem el punt com si fos l'últim.
      g.stroke (0, 0, 0);
      g.fill(255, 0, 255);
      pintaPunt(4, 4);
      //Actulitzem el valor de les etiquetes.
      controlador.resetValorsLabelsLateral();
      //Demanem actualitzar el panell de dades.
      controladorPanel.resetlabelsGravetat();
      controladorPanel.resetValorsCalibrat();
      //si el punter i el dial coincideixen, el pintem com si fos l'últim
    }
    //l'ultim el pintem màgenta.
    if (model.xDial == model.xUltim) {
      g.fill(255, 0, 255);
      g.stroke (255, 0, 255);
      pintaPunt(4, 4);
    } else {//sinó, el pintem normal.
      g.stroke (0, 0, 0);
      g.fill(0, 0, 0);
      pintaPunt(1, 1);
    }
  }
  
  /**
   Mètode per actualitzar la vista amb els valors de l'últim registre.
   Estem visualitzant les dades d'un arxiu i estem en pause.
   */
  private void pintaUltim() {
      model.xDial =model.xUltim;
      controlador.getUltimRegistre();
      g.stroke (0, 0, 0);
      g.fill(255, 0, 255);
      pintaPunt(4, 4);
      controlador.resetValorsLabelsLateral();
      controladorPanel.resetlabelsGravetat();
      controladorPanel.resetValorsCalibrat();
  }


  private void pintaPunt(int d1, int d2) {
    float y = controlador.getYPuntZonaSuperior();
    if ( controlador.isSeleccionatSuperior()) { 
      //Calcula la mitja des del dial clicat fins l'ultim (positius o negatius, no barreja) i la pinta.
      //  float mitja =controlador.getYMitjaZonaSuperior();
      // float yy = map(mitja, 0, model.roll_e_escala, BD.y_ZSuperiorZero, BD.y_ZSuperiorBottom);
      // g.stroke (255, 0, 0);
      // g.text(mitja, BD.x_fi_dials +6, yy);
      // g.line(model.xDial, yy, BD.x_fi_dials, yy);
      g.stroke (0, 0, 0);
      g.line(model.xDial, y, BD.x_fi_dials, y);
      g.text(model.getValorSuperior(), BD.x_fi_dials +2, y);
      g.ellipse(model.xDial, y, d1, d2);

      if ((model.idRegistre_pintar+1) % 4 !=0) {
        g.stroke (0, 0, 0);
        g.line(model.xDial, BD.y_dialTop+5, model.xDial, BD.y_dialBottom+4);
      } else {
        g.stroke (255, 0, 0);
        g.line(model.xDial, BD.y_dialTop, model.xDial, BD.y_dialBottom+4);
      }
    } else  g.ellipse(model.xDial, y, d1, d2);

    y = controlador.getYPuntZonaCentral();
    if ( controlador.isSeleccionatCentre()) { 
      g.ellipse(model.xDial, y, 3, 3);
      g.line(model.xDial, y, BD.x_fi_dials, y );
      g.text(model.getValorCentre(), BD.x_fi_dials +2, y);
      g.stroke (0, 0, 0);
      g.line(model.xDial, BD.y_dialTop+5, model.xDial, BD.y_dialBottom+4);
    } else  g.ellipse(model.xDial, y, d1, d2);

    y =  controlador.getYPuntZonaInferior();
    //Si cliquem el punt de la zona inferior, pintem ala dreta del gàfic el valor i la recta punt-valor.
    if ( controlador.isSeleccionatInferior()) { 
      g.ellipse(model.xDial, y, 3, 3);
      g.line(model.xDial, y, BD.x_fi_dials, y ); 
      g.text(model.getValorInferior(), BD.x_fi_dials +2, y);
      g.stroke (0, 0, 0);
      g.line(model.xDial, BD.y_dialTop+5, model.xDial, BD.y_dialBottom+4);
    } else  g.ellipse(model.xDial, y, d1, d2);
  }

  ////////////////////////////////////////////////////////////////////////////////////
  private void creaBotons() {
    //Instanciem el boto play/pause.
    int x = BD.left +5;
    int y = BD.top + 310;
    b_playPause= controlP5.addButton(BD.s_nom_play)
      .setId(128)
      .setPosition(x, y)
      .setImage(loadImage(BD.s_img_pause))
      .setSize(20, 20);

    //Afegim els listeners i les accions.
    b_playPause.addCallback(new CallbackListener() {
      public void controlEvent(CallbackEvent e) {
        switch(e.getAction()) {
          case(ControlP5.ACTION_ENTER): 
          if (model.pause) {
            b_playPause.setImage(loadImage(BD.s_img_play));
            loop();
          } else b_playPause.setImage(loadImage(BD.s_img_pause1));
          break;
          case(ControlP5.ACTION_LEAVE): 
          if (model.pause) {
            b_playPause.setImage(loadImage(BD.s_img_play1));
            loop();
          } else b_playPause.setImage(loadImage(BD.s_img_pause));
          break;
          case(ControlP5.ACTION_RELEASE): 
          if (model.pause) {
            b_playPause.setImage(loadImage(BD.s_img_pause));
            loop();
          } else  b_playPause.setImage(loadImage(BD.s_img_play));
          model.pause = !model.pause;
          break;
        }
      }
    }
    );
    //Instanciem la roda per als frames.
    x = BD.left + 30;
    y = BD.top + 315;
    r_frame=  new ControladorRoda(x, y, BD.w_roda, BD.h_roda, BD.idr_fps, BD.noms_r[BD.idr_fps]);
    r_frame.setRang(1, BD.fremsps, BD.s_fps);
    r_frame.setValor(model.v_fotograma, BD.s_fps);

    //Instanciem la llista desplegable per als arxius.
    ls= controlP5.addScrollableList(BD.NOM_LLISTA_ARXIUS)
      .setPosition(BD.x_llista, BD.y_llista)
      .setSize(BD.w_llista, BD.h_llista)
      .setBarHeight(BD.h_bar_llista)
      .setItemHeight(BD.h_item_llista)
      .setColorActive(0xffffffff)        //color de l'item al clicar.
      .setColorBackground(0xffff851B)    //color de fons
      .setColorLabel(0xff000000)         //Color lletra capsalera 
      .setColorValue(0xff000000)         //color de la lletra dels itmes.
      .setColorForeground(0x55ff851B)    //color de fons al pasar el punter.
      .setFont(createFont("Verdana", 11))
      .setId(BD.Id_llista)
      .addItems(BD.llistaArxius);
  }
  public String getValorLabel() {
    return ls.getCaptionLabel().getText();
  }
  public ControladorRoda getRodaVelocitat() {
    return r_frame;
  }
}
