//LLIBRERIES//////////////////////////////////////////////////////////
//Per a la comunicació I2C
#include <Wire.h>
//Per a llegir el Mòdul IMU
#include <Adafruit_BNO055.h>
//Per a treballar amb la tarjeta SD
#include <SD.h>
//Per a trebllar amb el Mòdul RTC
#include "RTClib.h"

//VARIABLES//////////////////////////////////////////////////////////+
//Volem capturar les dades cada 250 mil·lisegons
#define TEMPS_REGISTRE (250)
//Instanciem un objecte IMU amb id = 55 a l'adreça 0x28
Adafruit_BNO055 bno = Adafruit_BNO055(55, 0x28);
//Instanciem un objecte RTC i les variables per a controlar el temps.
RTC_DS3231 rtc;
boolean instalarHora = false;
boolean compensarHora = false;
//Cada 4 cicles de loop() verifiquem l'estat del calibrat i de la targeta.
int nCicles = 0;
//Variables per a controlar l'encesa del led vermell i verd.
const int pin_ledVerd = 8;     // LED connectat al pin 8
const int pin_ledVermell = 13; // LED connectat al pin 13
//Boolean per a definir l'estat del calibrat.
boolean calibrat = false;
//Boolean per a triar entre Monitor Serie o la targeta,
//per defecte enviem els registres sempre a la targeta.
boolean activatPortSerial = true;
//Arxiu per a recollir els registres.
//Cal activar una línia diferent per a cada dispositiu.
File arxiu;
//const String nomArxiu = "cintura.csv";
//const String nomArxiu = "cuixa.csv";
const String nomArxiu = "cama.csv";
///////////////////////////////////

//Funció setup, s'executa a l'iniciar el programa.
void setup() {
  //Cas d'ús: Iniciar programa.
  //Iniciem la comunicació I2C.
  Wire.begin();
  //Definim els pin 13 i 8 a OUTPUT.
  pinMode(pin_ledVermell, OUTPUT);
  pinMode(pin_ledVerd, OUTPUT);
  //Encenem el led vermell.
  digitalWrite(pin_ledVermell, HIGH);
  //Cas d'ús: Comprovar connexió amb tots els components.
  //Verifiquem que tenim connexió amb tots els mòduls del dispositiu
  if (!bno.begin() || !SD.begin(4) || !rtc.begin()) {
    //Cas d'ús: Comprovar que la targeta està connectada.
    //Verifiquem el que és més probable.
    testSD(100);
  } //Si tenim connexió amb tots els mòduls, el led vermell s’apagarà.
  else digitalWrite(pin_ledVermell, LOW);
  //Iniciem el monitor Serie
  Serial.begin(115200);
}
// Funció loop s'executa contínuament.
void loop() {
  //Controlem el temps que ocupa la rutina per tal de restar-lo del Delay.
  unsigned long tempsRutina = millis();
  //Mirem si tenim dades en el port serial.
  if (Serial.available() > 0) {
    char dada = Serial.read();
    //Cas d'ús: Instal·lar l'hora
    if ( dada == 'h') setHora();
    //Si estem instal·lant l'hora, no podem fer res més.
    else if (instalarHora)  return;
    //Cas d'ús: compensar l'hora.
    if ( dada == 't') setFlagCompensarHora();
    if (compensarHora) {
      ajustarHora(dada);
      //Si estem compensant l'hora, no podem fer res més.
      return;
    }
    //Cas d'ús: establir el mode de capturar el registre.
    if ( dada == 'p' )setModeRegistre();
  }
  //Cas d'ús: Estem instal·lant l'hora
  //Cas d'ús: Estem compensant l'hora.
  //No fem res mentre estem instal·lant o compensant l'hora
  if (compensarHora || instalarHora) return;
  //Cas d'ús: Construir un registre.
  if (activatPortSerial)  creaRegistreSerial();
  else creaRegistreSD();
  //Cas d'ús: ha transcorregut un segon. Cada segon, realitzem les rutines programades
  if (nCicles == 4) {
    nCicles = 0;
    //Cas d'ús: Comprovar l'estat del calibrat.
    avisCalibracio();
    //Cas d'ús: Comprovar que la targeta està connectada.
    testSD(2000);
  }
  else nCicles++;
  //Cas d'ús: Esperar 250 mil·lisegons.
  delay(TEMPS_REGISTRE - (millis() - tempsRutina) % TEMPS_REGISTRE );
}
/**
   Mètode per a posar l'hora del PC en el mòdul RTC
*/
void setHora() {
  //Iniciem el procés
  instalarHora = true;
  //Si havíem iniciat el procés de compensarà l'hora, cal aturar-lo.
  //primer cal instal·lar i desprès compensar.
  if (compensarHora) compensarHora = false;
  Serial.println("ACTIVAR INSTAL·LAR HORA ********");
  //Imprimim l'ora actual del RTC.
  imprimirHora();
  Serial.println(" <-- Hora dispositiu");
  //Llegim la cadena enviada.
  String str = Serial.readStringUntil('\n');
  //Eliminem els blancs dels extrems.
  str.trim();
  Serial.println("Cadena llegida: " + str);
  //Nomes controlem: tenir com a mínim 5 dígits i  tenir un caràcter ':'
  //després del/s dígits de l'hora.
  if (str.length() < 5 || str.indexOf(':') < 1 || str.indexOf(':') > 2) {
    Serial.println("Cadena amb errors !!!!!!!");
    Serial.println("Exemples correctes: h 14:22:33  o  h 4:2:3  o  h 14:2:33");
    Serial.println("TORNA A INTENTAR-HO !!!!! ********");
    return;
  }
  Serial.println("INSTAL·LANT HORA ********");
  uint8_t h, m, s = 0;
  for (int i = 0 ; i < 3; i++) {
    int id = str.indexOf(':');
    //Amb l'operació mod controlem els valors adients.
    if (i == 0) h = str.substring(0, id).toInt() % 24;
    else if (i == 1) m = str.substring(0, id).toInt() % 60;
    else if (i == 2) s = str.substring(0, id).toInt() % 60;
    str = str.substring(id + 1);
  }
  //Primer, carreguem la data i l'hora del sistema
  DateTime avui = DateTime(F(__DATE__), F(__TIME__));
  //Ara, modifiquem l'hora amb les dades rebudes.
  rtc.adjust(DateTime(avui.year(), avui.month(), avui.day(), h, m, s ));
  // Imprimim l'hora
  imprimirHora();
  Serial.println(" <-- Hora Instal·lada");
  Serial.println("FINAL INSTAL·LAR HORA ********");
  Serial.println("Envia 't' per a iniciar la rutina de compensar ********");
  Serial.println("CONTINUEM ENVIANT REGISTRES !!! ********");
  Serial.flush();
  instalarHora = false;
}

/**
   Mètode per a canviar el valor del boolean que permet iniciar la rutina
   per a fer la compensació de l'hora que té el mòdul RTC.
*/
void setFlagCompensarHora() {
  //Canviem l'estat del boolean.
  compensarHora = !compensarHora;
  //Donem el missatge adient.
  if (compensarHora) {
    Serial.println("ACTIVAR COMPENSAR HORA ********");
    Serial.println("ATUREM ELS REGISTRES !!! ********");
    Serial.println("Envia '+' o '-' per a ajustar els segons. ********");
    Serial.println("Envia 't' per a sortir de la rutina ********");
  }
  else {
    Serial.println("DESACTIVAR COMPENSAR HORA ********");
    Serial.println("Envia 't' per a iniciar la rutina ********");
    Serial.println("CONTINUEM ENVIANT REGISTRES !!! ********");
  }
  //Imprimim l'ora actual del RTC.
  imprimirHora();
  Serial.println(" <-- Hora dispositiu");
  Serial.flush();
}
/**
   Mètode per a compensar l'hora del RTC.
*/
void ajustarHora(char dada) {
  if ( dada == '+' || dada == '-' ) {
    DateTime ara = rtc.now();
    int s = ara.second();
    if (dada == '+') s++;
    else if (dada == '-')s--;
    if (s < 0) s = 0;
    else if (s > 59) s = 59;
    //Carreguem de nou l'hora. Només modifiquem els segons.
    rtc.adjust(DateTime(ara.year(), ara.month(), ara.day(), ara.hour(), ara.minute(), s));
    //Mostrem l'hora amb la compensació.
    imprimirHora();
    Serial.println(" <-- Hora Modificada");
    Serial.flush();
  }
}
/**
   Mètode per a encendre o apagar el led verd que informa de l'estat del calibrat
   Led verd encès = calibrat incorrecte.
   Led verd apagat = calibrat correcte.
*/
void avisCalibracio() {
  //Si el led verd està encès, el calibrat no és complet.
  if (calibrat) digitalWrite(pin_ledVerd, LOW);
  else  digitalWrite(pin_ledVerd, HIGH);
}
/**
   Mètode per a escriure l'hora en el Port Serial.
   hh:mm:ss o hh:m:s
*/
void imprimirHora() {
  DateTime now = rtc.now();
  Serial.print(now.hour(), DEC); Serial.print(':');
  Serial.print(now.minute(), DEC); Serial.print(':');
  Serial.print(now.second(), DEC); Serial.print(',');
}
/**
   Mètode per a canviar l'estat del boolean que controla si el registre
   S'escriu en el port Serial o a la targeta SD.
*/
void setModeRegistre() {
  //Canviem l'estat del boolean.
  activatPortSerial = !activatPortSerial;
  //Imprimim el missatge adient.
  if (activatPortSerial) {
    Serial.println("ACTIVAT MONITOR SERIE ********");
    Serial.println("Envia una 'p' per a enviar-los a la targeta SD ********");
  }
  else {
    Serial.println("ENVIANT REGISTRES A LA TARGETA SD ********");
    Serial.println("Envia una 'p' per a mostrar-los pel Monitor ********");
  }
  Serial.flush();
}
/**
   Mètode per a comprovar l'estat de la connexió de la targeta SD
   i encendre el led vermell.
   Led vermell fa pampallugues = La connexió no és correcta.
*/
void testSD(int tempsAturada) {
  //Pot passar que no haguem posat la targeta o que no faci bon contacte
  //en aquests casos el led vermell farà pampallugues.
  //tempsAturada = 2000 si detectem l'error al mètode loop()
  //tempsAturada = 100 si detectem l'error en el mètode setup()
  while (!SD.begin(4)) {
    digitalWrite(pin_ledVermell, HIGH);
    delay(tempsAturada);
    digitalWrite(pin_ledVermell, LOW);
  }
}
/**
   Mètode per a construïr el Registre per a mostrar-lo
   en el Monitor Serie.
*/
void creaRegistreSerial() {
  //INICI DEL REGISTRE.//////////////////////////////////////////////////////////////////
  //Inici del registre hh:mm:ss,
  imprimirHora();
  //Capturem els valors Yaw, Roll i Pitch
  imu::Vector<3> vec = bno.getVector(Adafruit_BNO055::VECTOR_EULER);
  Serial.print(vec.x()); Serial.print(',');//yaw: rotació de L'eix vertical.
  Serial.print(vec.y()); Serial.print(',');//roll: inclinació ESQUERRA/DRETA
  Serial.print(vec.z()); Serial.print(',');//pitch: inclinació DAVANT/DARRERE
  //Capturem el valor de les components del vector Gravetat.
  vec = bno.getVector(Adafruit_BNO055::VECTOR_GRAVITY);
  Serial.print(vec.x()); Serial.print(',');//gx
  Serial.print(vec.y()); Serial.print(',');//gy
  Serial.print(vec.z()); Serial.print(','); //gz
  //Capturem el valor de les components de l'acceleració
  vec = bno.getVector(Adafruit_BNO055::VECTOR_ACCELEROMETER);
  Serial.print(vec.x()); Serial.print(',');//gx
  Serial.print(vec.y()); Serial.print(',');//gy
  Serial.print(vec.z()); Serial.print(','); //gz
  //Capturem el resultat del calibrat.
  uint8_t sys, gyro, accel, mag = 0;
  bno.getCalibration(&sys, &gyro, &accel, &mag);
  calibrat = (gyro == 3 && accel == 3 && mag == 3 && sys == 3 );
  Serial.print(gyro, DEC); Serial.print(',');
  Serial.print(accel, DEC); Serial.print(',');
  Serial.print(mag, DEC); Serial.print(',');
  Serial.println(sys, DEC);
  //FI DEL REGISTRE.//////////////////////////////////////////////////////////////////////
}
/**
   Mètode per a construir un registre i escriure'l en
   la targeta SD.
*/
void creaRegistreSD() {
  //Si l'arxiu no existeix, el crea en la targeta.
  arxiu = SD.open(nomArxiu, FILE_WRITE);
  //INICI DEL REGISTRE.///////////////////////////////////////////////////////////////////
  //Capturem l'hora
  DateTime now = rtc.now();
  //Inici del registre hh:mm:ss,
  arxiu.print(now.hour(), DEC); arxiu.print(':');
  arxiu.print(now.minute(), DEC); arxiu.print(':');
  arxiu.print(now.second(), DEC); arxiu.print(',');
  //Capturem els valors Yaw, Roll i Pitch
  imu::Vector<3> vec = bno.getVector(Adafruit_BNO055::VECTOR_EULER);
  arxiu.print(vec.x()); arxiu.print(',');//yaw: rotació de l'eix vertical.
  arxiu.print(vec.y()); arxiu.print(',');//roll: inclinació ESQUERRA/DRETA
  arxiu.print(vec.z()); arxiu.print(',');//pitch: inclinació DAVANT/DARRERE
  //Capturem el valor de les components del vector Gravetat.
  vec = bno.getVector(Adafruit_BNO055::VECTOR_GRAVITY);
  arxiu.print(vec.x()); arxiu.print(',');
  arxiu.print(vec.y()); arxiu.print(',');
  arxiu.print(vec.z()); arxiu.print(',');
  //Capturem el valor de les components de l'acceleració
  vec = bno.getVector(Adafruit_BNO055::VECTOR_ACCELEROMETER);
  arxiu.print(vec.x()); arxiu.print(',');
  arxiu.print(vec.y()); arxiu.print(',');
  arxiu.print(vec.z()); arxiu.print(',');
  //Capturem el resultat del calibrat.
  uint8_t sys, gyro, accel, mag = 0;
  bno.getCalibration(&sys, &gyro, &accel, &mag);
  calibrat = (gyro == 3 && accel == 3 && mag == 3 && sys == 3 );
  arxiu.print(gyro, DEC); arxiu.print(',');
  arxiu.print(accel, DEC); arxiu.print(',');
  arxiu.print(mag, DEC); arxiu.print(',');
  arxiu.println(sys, DEC);
  //FI DEL REGISTRE.//////////////////////////////////////////////////////////////////////
  arxiu.close();
}
